# coding=utf-8
__author__ = "Dimitris Karkalousos"

import math
from typing import Optional

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import check_stacked_complex, coil_combination_method, rnn_weights_init
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.recurrentvarnet_base.recurrentvarnet_block import (
    RecurrentInit,
    RecurrentVarNetBlock,
)
from atommic.core.classes.common import typecheck

__all__ = ["RecurrentVarNet"]


class RecurrentVarNet(BaseMRIReconstructionModel):
    """Implementation of the Recurrent Variational Network implementation, as presented in [Yiasemis2021]_.

    References
    ----------
    .. [Yiasemis2021] Yiasemis, George, et al. “Recurrent Variational Network: A Deep Learning Inverse Problem Solver
        Applied to the Task of Accelerated MRI Reconstruction.” ArXiv:2111.09639 [Physics], Nov. 2021. arXiv.org,
        http://arxiv.org/abs/2111.09639.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`RecurrentVarNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.in_channels = cfg_dict.get("in_channels")
        self.recurrent_hidden_channels = cfg_dict.get("recurrent_hidden_channels")
        self.recurrent_num_layers = cfg_dict.get("recurrent_num_layers")
        self.no_parameter_sharing = cfg_dict.get("no_parameter_sharing")

        # make time-steps size divisible by 8 for fast fp16 training
        self.num_steps = 8 * math.ceil(cfg_dict.get("num_steps") / 8)

        self.learned_initializer = cfg_dict.get("learned_initializer")
        self.initializer_initialization = cfg_dict.get("initializer_initialization")
        self.initializer_channels = cfg_dict.get("initializer_channels")
        self.initializer_dilations = cfg_dict.get("initializer_dilations")

        if (
            self.learned_initializer
            and self.initializer_initialization is not None
            and self.initializer_channels is not None
            and self.initializer_dilations is not None
        ):
            if self.initializer_initialization not in [
                "sense",
                "input_image",
                "zero_filled",
            ]:
                raise ValueError(
                    "Unknown initializer_initialization. Expected `sense`, `'input_image` or `zero_filled`."
                    f"Got {self.initializer_initialization}."
                )
            self.initializer = RecurrentInit(
                self.in_channels,
                self.recurrent_hidden_channels,
                channels=self.initializer_channels,
                dilations=self.initializer_dilations,
                depth=self.recurrent_num_layers,
                multiscale_depth=cfg_dict.get("initializer_multiscale"),
            )
        else:
            self.initializer = None  # type: ignore

        self.block_list: torch.nn.Module = torch.nn.ModuleList()
        for _ in range(self.num_steps if self.no_parameter_sharing else 1):
            self.block_list.append(
                RecurrentVarNetBlock(
                    in_channels=self.in_channels,
                    hidden_channels=self.recurrent_hidden_channels,
                    num_layers=self.recurrent_num_layers,
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                    coil_dim=self.coil_dim,
                )
            )

        std_init_range = 1 / self.recurrent_hidden_channels**0.5

        # initialize weights if not using pretrained cirim
        if not cfg_dict.get("pretrained", False):
            self.block_list.apply(lambda module: rnn_weights_init(module, std_init_range))

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`RecurrentVarNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        previous_state: Optional[torch.Tensor] = None

        if self.initializer is not None:
            if self.initializer_initialization in ("sense", "input_image"):
                initializer_input_image = initial_prediction.unsqueeze(self.coil_dim)
            elif self.initializer_initialization == "zero_filled":
                initializer_input_image = ifft2(
                    y,
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                )

            previous_state = self.initializer(
                fft2(
                    initializer_input_image,
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                )
                .sum(1)
                .permute(0, 3, 1, 2)
            )

        kspace_prediction = y.clone()
        for step in range(self.num_steps):
            block = self.block_list[step] if self.no_parameter_sharing else self.block_list[0]
            kspace_prediction, previous_state = block(
                kspace_prediction,
                y,
                mask,
                sensitivity_maps,
                previous_state,
            )

        return check_stacked_complex(
            coil_combination_method(
                ifft2(kspace_prediction, self.fft_centered, self.fft_normalization, self.spatial_dims),
                sensitivity_maps,
                self.coil_combination_method,
                self.coil_dim,
            )
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import argparse

from atommic.cli.launch import register_cli_subcommand


def main():
    """Run the CLI."""
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    subparser = parser.add_subparsers(help="atommic commands.")
    subparser.required = True
    subparser.dest = "subcommand"

    register_cli_subcommand(subparser)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from omegaconf import DictConfig

from atommic.collections.segmentation.nn.attentionunet_base.attentionunet_block import AttentionUnet
from atommic.collections.segmentation.nn.segmentationnet import BaseSegmentationNet

__all__ = ["SegmentationAttentionUNet"]


# type: ignore
class SegmentationAttentionUNet(BaseSegmentationNet):
    """Implementation of the Attention UNet for MRI segmentation, as presented in [Oktay2018]_.

    References
    ----------
    .. [Oktay2018] O. Oktay, J. Schlemper, L.L. Folgoc, M. Lee, M. Heinrich, K. Misawa, K. Mori, S. McDonagh, N.Y.
        Hammerla, B. Kainz, B. Glocker, D. Rueckert. Attention U-Net: Learning Where to Look for the Pancreas. 2018.
        https://arxiv.org/abs/1804.03999

    """

    def build_segmentation_module(self, cfg: DictConfig) -> torch.nn.Module:
        """Build the segmentation module.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object specifying the model's hyperparameters.

        Returns
        -------
        torch.nn.Module
            The segmentation module.
        """
        return AttentionUnet(
            in_chans=self.input_channels,
            out_chans=cfg.get("segmentation_module_output_channels", 2),
            chans=cfg.get("segmentation_module_channels", 64),
            num_pool_layers=cfg.get("segmentation_module_pooling_layers", 2),
            drop_prob=cfg.get("segmentation_module_dropout", 0.0),
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/decorators/deprecated.py

__all__ = ["deprecated"]

import functools
import inspect
from typing import Dict

import wrapt

# Remember which deprecation warnings have been printed already.
from atommic.utils import logging

_PRINTED_WARNING: Dict = {}


def deprecated(wrapped=None, version=None, explanation=None):
    """This is a decorator which can be used to mark functions as deprecated. It will result in a warning being emitted
    when the function is used.

    Parameters
    ----------
    wrapped : function
        The function to be decorated.
    version : str
        The version of the package where the function was deprecated.
    explanation : str
        The explanation of the deprecation.

    Returns
    -------
    function
        The decorated function.
    """
    if wrapped is None:
        return functools.partial(deprecated, version=version, explanation=explanation)

    @wrapt.decorator
    def wrapper(_wrapped, args, kwargs):
        """Prints the adequate warning (only once per function) when required and calls the function func, passing the
        original arguments, i.e. version and explanation.

        Parameters
        ----------
        _wrapped : function
            The function to be decorated.
        args : tuple
            The arguments passed to the function to be decorated.
        kwargs : dict
            The keyword arguments passed to the function to be decorated.

        Returns
        -------
        function
            The decorated function.
        """
        # Check if we already warned about that function.
        if _wrapped.__name__ not in _PRINTED_WARNING:
            # Add to list, so we won't print it again.
            _PRINTED_WARNING[_wrapped.__name__] = True

            # Prepare the warning message.
            entity_name = "Class" if inspect.isclass(wrapped) else "Function"
            msg = f"{entity_name} '{_wrapped.__name__}' is deprecated."

            # Optionally, add version and explanation.
            if version is not None:
                msg = f"{msg} It is going to be removed in the {version} version."

            if explanation is not None:
                msg = f"{msg} {explanation}"

            # Display the deprecated warning.
            logging.warning(msg)

        # Call the function.
        return _wrapped(*args, **kwargs)

    return wrapper(wrapped)  # pylint: disable=no-value-for-parameter



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from torch import nn


class CrossEntropyLoss(nn.Module):
    """Wrapper around PyTorch's CrossEntropyLoss to support 2D and 3D inputs."""

    def __init__(
        self,
        num_samples: int = 50,
        ignore_index: int = -100,
        reduction: str = "none",
        label_smoothing: float = 0.0,
        weight: torch.Tensor = None,
    ):
        """Inits :class:`CrossEntropyLoss`.

        Parameters
        ----------
        num_samples : int, optional
            Number of Monte Carlo samples, by default 50
        ignore_index : int, optional
            Index to ignore, by default -100
        reduction : str, optional
            Reduction method, by default "none"
        label_smoothing : float, optional
            Label smoothing, by default 0.0
        weight : torch.Tensor, optional
            Weight for each class, by default None
        """
        super().__init__()
        self.mc_samples = num_samples
        self.cross_entropy = torch.nn.CrossEntropyLoss(
            weight=weight,
            ignore_index=ignore_index,
            reduction=reduction,
            label_smoothing=label_smoothing,
        )

    def forward(self, target: torch.Tensor, _input: torch.Tensor, pred_log_var: torch.Tensor = None) -> torch.Tensor:
        """Forward pass of :class:`CrossEntropyLoss`.

        Parameters
        ----------
        target : torch.Tensor
            Target tensor. Shape: (batch_size, num_classes, *spatial_dims)
        _input : torch.Tensor
            Prediction tensor. Shape: (batch_size, num_classes, *spatial_dims)
        pred_log_var : torch.Tensor, optional
            Prediction log variance tensor. Shape: (batch_size, num_classes, *spatial_dims). Default is ``None``.

        Returns
        -------
        torch.Tensor
            Loss tensor. Shape: (batch_size, *spatial_dims)
        """
        # In case we do not have a batch dimension, add it
        if _input.dim() == 3:
            _input = _input.unsqueeze(0)
        if target.dim() == 3:
            target = target.unsqueeze(0)

        self.cross_entropy.weight = self.cross_entropy.weight.clone().to(_input.device)

        if self.mc_samples == 1 or pred_log_var is None:
            return self.cross_entropy(_input.float(), target).mean()

        pred_shape = [self.mc_samples, *_input.shape]
        noise = torch.randn(pred_shape, device=_input.device)
        noisy_pred = _input.unsqueeze(0) + torch.sqrt(torch.exp(pred_log_var)).unsqueeze(0) * noise
        noisy_pred = noisy_pred.view(-1, *_input.shape[1:])
        tiled_target = target.unsqueeze(0).tile((self.mc_samples,)).view(-1, *target.shape[1:])
        loss = self.cross_entropy(noisy_pred, tiled_target).view(self.mc_samples, -1, *_input.shape[-2:]).mean(0)
        return loss.mean()



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/collections/common/parts/patch_utils.py

from packaging import version

# Library version globals
TORCH_VERSION = None
TORCH_VERSION_MIN = version.Version("1.12.0")



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/utils/neural_type_utils.py

from collections import defaultdict


def get_io_names(types, disabled_names):
    """This method will return a list of input and output names for a given NeuralType.

    Parameters
    ----------
    types : type
        The NeuralType of the module or model to be inspected.
    disabled_names : list
        A list of names that should be excluded from the result.

    Returns
    -------
    list
        A list of input and output names.
    """
    names = list(types.keys())
    for name in disabled_names:
        if name in names:
            names.remove(name)
    return names


def get_dynamic_axes(types, names):
    """This method will return a dictionary with input/output names as keys and a list of dynamic axes as values.

    Parameters
    ----------
    types : NeuralType
        The NeuralType of the module or model to be inspected.
    names : list
        A list of names that should be inspected.

    Returns
    -------
    dict
        A dictionary with input/output names as keys and a list of dynamic axes as values.
    """
    dynamic_axes = defaultdict(list)
    if names is not None:
        for name in names:
            if name in types:
                dynamic_axes.update(extract_dynamic_axes(name, types[name]))  # noqa: F821
    return dynamic_axes



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/env_var_parsing.py

import decimal
import json
import os

from dateutil import parser  # type: ignore

__all__ = [
    "get_env",
    "get_envbool",
    "get_envint",
    "get_envfloat",
    "get_envdecimal",
    "get_envdate",
    "get_envdatetime",
    "get_envlist",
    "get_envdict",
    "CoercionError",
    "RequiredSettingMissingError",
]


class CoercionError(Exception):
    """Custom error raised when a value cannot be coerced."""

    def __init__(self, key, value, func):
        """Inits :class:`CoercionError`.

        Parameters
        ----------
        key : str
            The name of the env var that is missing.
        value : str
            The value of the env var that is missing.
        func : function
            The function that was used to try and coerce the value.
        """
        msg = f"Unable to coerce '{key}={value}' using {func.__name__}."
        super().__init__(msg)


class RequiredSettingMissingError(Exception):
    """Custom error raised when a required env var is missing."""

    def __init__(self, key):
        """Inits :class:`RequiredSettingMissingError`.

        Parameters
        ----------
        key : str
            The name of the env var that is missing.
        """
        msg = f"Required env var '{key}' is missing."
        super().__init__(msg)


def _get_env(key, default=None, coerce=lambda x: x, required=False):
    """Return env var coerced into a type other than string. This function extends the standard os.getenv function to
    enable the coercion of values into data types other than string (all env vars are strings by default).

    Parameters
    ----------
    key : str
        The name of the env var to retrieve.
    default : str, (optional)
        The default value to return if the env var is not set. NB the default value is **not** coerced, and is assumed
        to be of the correct type.
    coerce : function, (optional)
        A function that takes a string and returns a value of the desired type.
    required : bool, (optional)
        If True, raises a RequiredSettingMissingError if the env var is not set.

    Returns
    -------
    str
        The value of the env var coerced into the desired type.
    """
    try:
        value = os.environ[key]
    except KeyError as e:
        if required is True:
            raise RequiredSettingMissingError(key) from e
        return default

    try:
        return coerce(value)
    except Exception as exc:
        raise CoercionError(key, value, coerce) from exc


# standard type coercion functions
def _bool(value):
    """Return env var cast as boolean."""
    if isinstance(value, bool):
        return value

    return value is not None and value.lower() not in (
        "false",
        "0",
        "no",
        "n",
        "f",
        "none",
    )


def _int(value):
    """Return env var cast as integer."""
    return int(value)


def _float(value):
    """Return env var cast as float."""
    return float(value)


def _decimal(value):
    """Return env var cast as Decimal."""
    return decimal.Decimal(value)


def _dict(value):
    """Return env var as a dict."""
    return json.loads(value)


def _datetime(value):
    """Return env var as a datetime."""
    return parser.parse(value)


def _date(value):
    """Return env var as a date."""
    return parser.parse(value).date()


def get_env(key, *default, **kwargs):
    """Return env var. This is the parent function of all other get_foo functions, and is responsible for unpacking
    args/kwargs into the values that _get_env expects (it is the root function that actually interacts with environ).

    Parameters
    ----------
    key : str
        The env var name to look up.
    default : str, (optional)
        The value to use if the env var does not exist. If this value is not supplied, then the env var is considered
        to be required, and a RequiredSettingMissingError error will be raised if it does not exist.
    kwargs : dict, (optional)
        Additional keyword arguments to pass to _get_env.

    Returns
    -------
    str
        The env var, coerced if required, and a default if supplied.
    """
    if len(default) not in (0, 1):
        raise AssertionError("Too many args supplied.")
    func = kwargs.get("coerce", lambda x: x)
    required = len(default) == 0
    default = None if required else default[0]
    return _get_env(key, default=default, coerce=func, required=required)


def get_envbool(key, *default):
    """Return env var cast as boolean."""
    return get_env(key, *default, coerce=_bool)


def get_envint(key, *default):
    """Return env var cast as integer."""
    return get_env(key, *default, coerce=_int)


def get_envfloat(key, *default):
    """Return env var cast as float."""
    return get_env(key, *default, coerce=_float)


def get_envdecimal(key, *default):
    """Return env var cast as Decimal."""
    return get_env(key, *default, coerce=_decimal)


def get_envdate(key, *default):
    """Return env var as a date."""
    return get_env(key, *default, coerce=_date)


def get_envdatetime(key, *default):
    """Return env var as a datetime."""
    return get_env(key, *default, coerce=_datetime)


def get_envlist(key, *default, **kwargs):
    """Return env var as a list."""
    separator = kwargs.get("separator", " ")
    return get_env(key, *default, coerce=lambda x: x.split(separator))


def get_envdict(key, *default):
    """Return env var as a dict."""
    return get_env(key, *default, coerce=_dict)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/nemo_logging.py

import enum
import logging as _logging
import sys
import threading
import warnings
from contextlib import contextmanager
from logging.handlers import MemoryHandler

from atommic.constants import ATOMMIC_ENV_VARNAME_REDIRECT_LOGS_TO_STDERR, ATOMMIC_ENV_VARNAME_TESTING
from atommic.utils.env_var_parsing import get_envbool
from atommic.utils.formaters.base import BaseATOMMICFormatter, DebugATOMMICFormatter
from atommic.utils.get_rank import is_global_rank_zero
from atommic.utils.metaclasses import Singleton

__all__ = ["Logger", "LogMode"]


class LogMode(enum.IntEnum):
    """Enum for the different logging modes."""

    EACH = 0  # Log the message each time
    # Log the message only once. The same message will not be logged again.
    ONCE = 1


class Logger(metaclass=Singleton):
    """Singleton class for logging."""

    # Level 0
    NOTSET = _logging.NOTSET

    # Level 10
    DEBUG = _logging.DEBUG

    # Level 20
    INFO = _logging.INFO

    # Level 30
    WARNING = _logging.WARNING

    # Level 40
    ERROR = _logging.ERROR

    # Level 50
    CRITICAL = _logging.CRITICAL

    _level_names = {0: "NOTSET", 10: "DEBUG", 20: "INFO", 30: "WARNING", 40: "ERROR", 50: "CRITICAL"}

    def __init__(self, capture_warnings=True):
        """Inits :class:`Logger`.

        Parameters
        ----------
        capture_warnings : bool
            If True, redirect all warnings to the logging package. If False, ensure that warnings are not redirected to
            logging but to their original destinations. Default is ``True``.
        """
        self._logger = None
        # Multi-GPU runs in separate processes, thread locks shouldn't be needed
        self._logger_lock = threading.Lock()
        self._handlers = {}
        self.old_warnings_showwarning = None
        self._define_logger(capture_warnings)
        self.once_logged = set()
        self.rank = 0 if is_global_rank_zero() else "UNK"

    # pylint: disable=inconsistent-return-statements
    def _define_logger(self, capture_warnings=True):
        """Creates the logger if not already created. Called in init."""
        # Use double-checked locking to avoid taking lock unnecessarily.
        if self._logger is not None:
            return self._logger

        with self._logger_lock:
            try:
                self._logger = _logging.getLogger("atommic_logger")
                # By default, silence all loggers except the logger for rank 0
                self.remove_stream_handlers()
                # If atommic_TESTING is set, add a streamhandler to all ranks
                if get_envbool(ATOMMIC_ENV_VARNAME_TESTING, False):
                    old_factory = _logging.getLogRecordFactory()

                    def record_factory(*args, **kwargs):
                        """Add rank to log records."""
                        record = old_factory(*args, **kwargs)
                        record.rank = self.rank
                        return record

                    _logging.setLogRecordFactory(record_factory)
                    self.add_stream_handlers(formatter=DebugATOMMICFormatter)
                elif is_global_rank_zero():
                    self.add_stream_handlers()

                # Add memoryhandlers, essentially buffers. They are used to save messages that we will flush to file
                # once the appropriate file handlers are added.
                if is_global_rank_zero():
                    # Add a memory handler for error messages. Only logged on rank 0
                    self._handlers["memory_err"] = MemoryHandler(-1)
                    self._handlers["memory_err"].addFilter(lambda record: record.levelno > _logging.INFO)
                    formatter = BaseATOMMICFormatter
                    self._handlers["memory_err"].setFormatter(formatter())
                    self._logger.addHandler(self._handlers["memory_err"])
                # Add a memory handler for all messages on all ranks
                self._handlers["memory_all"] = MemoryHandler(-1)
                formatter = BaseATOMMICFormatter
                self._handlers["memory_all"].setFormatter(formatter())
                self._logger.addHandler(self._handlers["memory_all"])

            finally:
                level = Logger.INFO
                if get_envbool(ATOMMIC_ENV_VARNAME_TESTING, False):
                    level = Logger.DEBUG
                self.set_verbosity(verbosity_level=level)
                self.captureWarnings(capture_warnings)

        self._logger.propagate = False

    def remove_stream_handlers(self):
        """Removes StreamHandler that log to stdout and stderr from the logger."""
        if self._logger is None:
            raise RuntimeError("Impossible to set handlers if the Logger is not predefined")

        # ======== Remove Handler if already existing ========
        try:
            self._logger.removeHandler(self._handlers["stream_stdout"])
            del self._handlers["stream_stdout"]
        except KeyError:
            pass

        try:
            self._logger.removeHandler(self._handlers["stream_stderr"])
            del self._handlers["stream_stderr"]
        except KeyError:
            pass

    def add_stream_handlers(self, formatter=BaseATOMMICFormatter):
        """Add StreamHandler that log to stdout and stderr to the logger. INFO and lower logs are streamed to stdout
        while WARNING and higher are streamed to stderr. If the ATOMMIC_ENV_VARNAME_REDIRECT_LOGS_TO_STDERR environment
        variable is set, all logs are sent to stderr instead.
        """
        if self._logger is None:
            raise RuntimeError("Impossible to set handlers if the Logger is not predefined")

        # Add the output handler.
        if get_envbool(ATOMMIC_ENV_VARNAME_REDIRECT_LOGS_TO_STDERR, False):
            self._handlers["stream_stdout"] = _logging.StreamHandler(sys.stderr)

        else:
            self._handlers["stream_stdout"] = _logging.StreamHandler(sys.stdout)
            self._handlers["stream_stdout"].addFilter(lambda record: record.levelno <= _logging.INFO)

            self._handlers["stream_stderr"] = _logging.StreamHandler(sys.stderr)
            self._handlers["stream_stderr"].addFilter(lambda record: record.levelno > _logging.INFO)

        self._handlers["stream_stdout"].setFormatter(formatter())
        self._logger.addHandler(self._handlers["stream_stdout"])

        try:
            self._handlers["stream_stderr"].setFormatter(formatter())
            self._logger.addHandler(self._handlers["stream_stderr"])
        except KeyError:
            pass

    def reset_stream_handler(self, formatter=BaseATOMMICFormatter):
        """Removes then adds stream handlers."""
        self.remove_stream_handlers()
        self.add_stream_handlers(formatter=formatter)

    def add_file_handler(self, log_file):
        """Add a FileHandler to logger that logs all messages to a file. If the logger had a MemoryHandler at
        self._handlers["memory_all"], those buffered messages are flushed to the new file, and the MemoryHandler is
        closed.
        """
        if self._logger is None:
            raise RuntimeError("Impossible to set handlers if the Logger is not predefined")

        self._handlers["file"] = _logging.FileHandler(log_file)
        formatter = BaseATOMMICFormatter
        self._handlers["file"].setFormatter(formatter())
        self._logger.addHandler(self._handlers["file"])

        if self._handlers.get("memory_all"):
            self._handlers["memory_all"].setTarget(self._handlers["file"])
            self._handlers["memory_all"].close()  # flush and remove
            del self._handlers["memory_all"]

    def add_err_file_handler(self, log_file):
        """Add a FileHandler to logger that logs all WARNING and higher messages to a file. If the logger had a
        MemoryHandler at self._handlers["memory_err"], those buffered messages are flushed to the new file, and the
        MemoryHandler is closed.
        """
        if self._logger is None:
            raise RuntimeError("Impossible to set handlers if the Logger is not predefined")

        self._handlers["file_err"] = _logging.FileHandler(log_file)
        self._handlers["file_err"].addFilter(lambda record: record.levelno > _logging.INFO)

        formatter = BaseATOMMICFormatter
        self._handlers["file_err"].setFormatter(formatter())
        self._logger.addHandler(self._handlers["file_err"])

        if self._handlers.get("memory_err"):
            self._handlers["memory_err"].setTarget(self._handlers["file_err"])
            self._handlers["memory_err"].close()  # flush and remove
            del self._handlers["memory_err"]

    # pylint: disable=inconsistent-return-statements
    def getEffectiveLevel(self):
        """Return how much logging output will be produced."""
        if self._logger is not None:
            return self._logger.getEffectiveLevel()

    def get_verbosity(self):
        """See getEffectiveLevel"""
        return self.getEffectiveLevel()

    def setLevel(self, verbosity_level):
        """Sets the threshold for what messages will be logged."""
        if self._logger is not None:
            self._logger.setLevel(verbosity_level)

            for handler in self._logger.handlers:
                handler.setLevel(verbosity_level)

    def set_verbosity(self, verbosity_level):
        """See setLevel"""
        self.setLevel(verbosity_level)

    @contextmanager
    def patch_stderr_handler(self, stream):
        """Sends messages that should log to stderr to stream instead. Useful for unittests"""
        if self._logger is None:
            raise RuntimeError("Impossible to patch logging handlers if handler does not exist")
        try:
            old_stream = self._handlers["stream_stderr"].stream
            if old_stream is None:
                raise ValueError

            # Port backwards set_stream() from python 3.7
            self._handlers["stream_stderr"].acquire()
            try:
                self._handlers["stream_stderr"].flush()
                self._handlers["stream_stderr"].stream = stream
            finally:
                self._handlers["stream_stderr"].release()

            yield stream
        except (KeyError, ValueError) as e:
            raise RuntimeError("Impossible to patch logging handlers if handler does not exist") from e

        finally:
            # Port backwards set_stream() from python 3.7
            self._handlers["stream_stderr"].acquire()
            try:
                self._handlers["stream_stderr"].flush()
                self._handlers["stream_stderr"].stream = old_stream
            finally:
                self._handlers["stream_stderr"].release()

    @contextmanager
    def patch_stdout_handler(self, stream):
        """Sends messages that should log to stdout to stream instead. Useful for unittests"""
        if self._logger is None:
            raise RuntimeError("Impossible to patch logging handlers if handler does not exist")
        try:
            old_stream = self._handlers["stream_stdout"].stream
            if old_stream is None:
                raise ValueError

            # Port backwards set_stream() from python 3.7
            self._handlers["stream_stdout"].acquire()
            try:
                self._handlers["stream_stdout"].flush()
                self._handlers["stream_stdout"].stream = stream
            finally:
                self._handlers["stream_stdout"].release()

            yield stream
        except (KeyError, ValueError) as e:
            raise RuntimeError("Impossible to patch logging handlers if handler does not exist") from e

        finally:
            # Port backwards set_stream() from python 3.7
            self._handlers["stream_stdout"].acquire()
            try:
                self._handlers["stream_stdout"].flush()
                self._handlers["stream_stdout"].stream = old_stream
            finally:
                self._handlers["stream_stdout"].release()

    @contextmanager
    def temp_verbosity(self, verbosity_level):
        """Sets a temporary threshold for what messages will be logged."""
        if self._logger is not None:
            old_verbosity = self.get_verbosity()

            try:
                self.set_verbosity(verbosity_level)
                yield

            finally:
                self.set_verbosity(old_verbosity)

        else:
            try:
                yield

            finally:
                pass

    def captureWarnings(self, capture):
        """If capture is true, redirect all warnings to the logging package. If capture is False, ensure that warnings
        are not redirected to logging but to their original destinations.
        """
        if self._logger is not None:
            if capture and self.old_warnings_showwarning is None:
                # Backup Method
                self.old_warnings_showwarning = warnings.showwarning
                warnings.showwarning = self._showwarning

            elif not capture and self.old_warnings_showwarning is not None:
                # Restore Method
                warnings.showwarning = self.old_warnings_showwarning
                self.old_warnings_showwarning = None

    def _showwarning(self, message, category, filename, lineno, file, line=None):  # pylint: disable=unused-argument
        """Implementation of show warnings which redirects to logging.
        It will call warnings.formatwarning and will log the resulting string with level logging.WARNING.
        """
        s = warnings.formatwarning(message, category, filename, lineno, line)
        self.warning("%s", s)

    def _logged_once(self, msg, mode):
        """Returns True if the given message has been logged at least once in the given mode.

        Parameters
        ----------
        msg : str
            The message to check.
        mode : LogMode
            The mode to check.

        Returns
        -------
        bool
            True if the message has been logged at least once in the given mode.
        """
        if mode == LogMode.ONCE:
            PREFIX_LEN = 12
            if msg[PREFIX_LEN:] in self.once_logged:
                return True
            self.once_logged.add(msg[PREFIX_LEN:])
        return False

    def debug(self, msg, *args, mode=LogMode.EACH, **kwargs):
        """Log 'msg % args' with severity 'DEBUG'.
        To pass exception information, use the keyword argument exc_info with a true value, e.g.
        logger.debug("Houston, we have %s", "thorny problem", exc_info=1)
        """
        if self._logger is not None and self._logger.isEnabledFor(Logger.DEBUG) and not self._logged_once(msg, mode):
            self._logger._log(Logger.DEBUG, msg, args, **kwargs)  # pylint: disable=protected-access

    def info(self, msg, *args, mode=LogMode.EACH, **kwargs):
        """Log 'msg % args' with severity 'INFO'.
        To pass exception information, use the keyword argument exc_info with a true value, e.g.
        logger.info("Houston, we have %s", "interesting problem", exc_info=1)
        """
        if self._logger is not None and self._logger.isEnabledFor(Logger.INFO) and not self._logged_once(msg, mode):
            self._logger._log(Logger.INFO, msg, args, **kwargs)  # pylint: disable=protected-access

    def warning(self, msg, *args, mode=LogMode.EACH, **kwargs):
        """Log 'msg % args' with severity 'WARNING'.
        To pass exception information, use the keyword argument exc_info with a true value, e.g.
        logger.warning("Houston, we have %s", "bit of a problem", exc_info=1)
        """
        if self._logger is not None and self._logger.isEnabledFor(Logger.WARNING) and not self._logged_once(msg, mode):
            self._logger._log(Logger.WARNING, msg, args, **kwargs)  # pylint: disable=protected-access

    def error(self, msg, *args, mode=LogMode.EACH, **kwargs):
        """Log 'msg % args' with severity 'ERROR'.
        To pass exception information, use the keyword argument exc_info with a true value, e.g.
        logger.error("Houston, we have %s", "major problem", exc_info=1)
        """
        if self._logger is not None and self._logger.isEnabledFor(Logger.ERROR) and not self._logged_once(msg, mode):
            self._logger._log(Logger.ERROR, msg, args, **kwargs)  # pylint: disable=protected-access

    def critical(self, msg, *args, mode=LogMode.EACH, **kwargs) -> None:
        """Log 'msg % args' with severity 'CRITICAL'.
        To pass exception information, use the keyword argument exc_info with a true value, e.g.
        logger.critical("Houston, we have %s", "major disaster", exc_info=1)

        Parameters
        ----------
        msg : str
            The message to log.
        *args : tuple
            The arguments to the message.
        mode : LogMode
            The mode to log the message in.
        **kwargs : dict
            The keyword arguments to the message.
        """
        if (
            self._logger is not None
            and self._logger.isEnabledFor(Logger.CRITICAL)
            and not self._logged_once(msg, mode)
        ):
            self._logger._log(Logger.CRITICAL, msg, args, **kwargs)  # pylint: disable=protected-access



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.multitask.rs.parts.transforms import RSMRIDataTransforms

__all__ = ["SegmentationMRIDataTransforms"]


class SegmentationMRIDataTransforms(RSMRIDataTransforms):
    """Transforms for the MRI segmentation task.

    .. note::
        Extends :class:`atommic.collections.multitask.rs.parts.transforms.RSMRIDataTransforms`.
    """



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import List, Optional, Tuple

import torch

from atommic.collections.common.parts import utils
from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.quantitative.nn.base import SignalForwardModel


class qVarNetBlock(torch.nn.Module):
    """Implementation of the quantitative End-to-end Variational Network (qVN), as presented in [Zhang2022]_.

    References
    ----------
    .. [Zhang2022] Zhang C, Karkalousos D, Bazin PL, Coolen BF, Vrenken H, Sonke JJ, Forstmann BU, Poot DH, Caan MW.
        A unified model for reconstruction and R2* mapping of accelerated 7T data using the quantitative recurrent
        inference machine. NeuroImage. 2022 Dec 1;264:119680.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
        no_dc: bool = False,
        linear_forward_model=None,
    ):
        """Inits :class:`qVarNetBlock`.

        Parameters
        ----------
        model : torch.nn.Module
            Model to apply soft data consistency.
        fft_centered : bool, optional
            Whether to center the fft. Default is ``False``.
        fft_normalization : str, optional
            The normalization of the fft. Default is ``backward``.
        spatial_dims : tuple, optional
            The spatial dimensions of the data. Default is ``None``.
        coil_dim : int, optional
            The dimension of the coils. Default is ``1``.
        no_dc : bool, optional
            Whether to not apply the DC component. Default is ``False``.
        linear_forward_model : torch.nn.Module, optional
            Linear forward model. Default is ``None``.
        """
        super().__init__()

        self.linear_forward_model = (
            SignalForwardModel(sequence="MEGRE") if linear_forward_model is None else linear_forward_model
        )

        self.model = model
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim
        self.no_dc = no_dc
        self.dc_weight = torch.nn.Parameter(torch.ones(1))

    def sens_expand(self, x: torch.Tensor, sens_maps: torch.Tensor) -> torch.Tensor:
        """Combines the sensitivity maps with coil-combined data to get multicoil data.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        sens_maps : torch.Tensor
            Coil Sensitivity maps.

        Returns
        -------
        torch.Tensor
            Expanded multicoil data.
        """
        return fft2(
            utils.complex_mul(x, sens_maps),
            centered=self.fft_centered,
            normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )

    def sens_reduce(self, x: torch.Tensor, sens_maps: torch.Tensor) -> torch.Tensor:
        """Combines the sensitivity maps with multicoil data to get coil-combined data.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        sens_maps : torch.Tensor
            Coil Sensitivity maps.

        Returns
        -------
        torch.Tensor
            SENSE coil-combined reconstruction.
        """
        x = ifft2(x, centered=self.fft_centered, normalization=self.fft_normalization, spatial_dims=self.spatial_dims)
        return utils.complex_mul(x, utils.complex_conj(sens_maps)).sum(dim=self.coil_dim)

    def forward(
        self,
        prediction: torch.Tensor,
        masked_kspace: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        sampling_mask: torch.Tensor,
        TEs: List,
    ) -> torch.Tensor:
        """Forward pass of :class:`qVarNetBlock`.

        Parameters
        ----------
        prediction : torch.Tensor
            Initial prediction of the quantitative maps.
        masked_kspace : torch.Tensor
            Subsampled k-space of shape [batch_size, n_coils, n_x, n_y, 2].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2].
        sampling_mask : torch.Tensor
            Sampling mask of shape [batch_size, 1, n_x, n_y, 1].
        TEs : List
            List of echo times.

        Returns
        -------
        torch.Tensor
            Reconstructed image of shape [batch_size, n_coils, n_x, n_y, 2].
        """
        initial_prediction = self.linear_forward_model(
            prediction[:, 0, ...].unsqueeze(0),  # R2*
            prediction[:, 1, ...].unsqueeze(0),  # S0
            prediction[:, 2, ...].unsqueeze(0),  # B0
            prediction[:, 3, ...].unsqueeze(0),  # phi
            TEs,
        )
        initial_prediction_kspace = self.sens_expand(initial_prediction, sensitivity_maps.unsqueeze(self.coil_dim - 1))
        soft_dc = (initial_prediction_kspace - masked_kspace) * sampling_mask * self.dc_weight
        initial_prediction = self.sens_reduce(soft_dc, sensitivity_maps.unsqueeze(self.coil_dim - 1)).to(masked_kspace)

        prediction = torch.view_as_real(prediction + torch.view_as_complex(self.model(initial_prediction)))
        prediction_tmp = prediction[:, 0, ...]
        prediction_tmp[prediction_tmp < 0] = 0
        prediction[:, 0, ...] = prediction_tmp

        return torch.abs(torch.view_as_complex(prediction))



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/lucidrains/lion-pytorch/tree/main/lion_pytorch

import torch
from torch.optim.optimizer import Optimizer

__all__ = ["Lion"]


class Lion(Optimizer):
    """Implements Lion, EvoLved Sign Momentum optimizer.

    This implementation is based on: `Symbolic Discovery of Optimization Algorithms` (see
    https://arxiv.org/abs/2302.06675)
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0):
        """Inits :class:`Lion`.

        Parameters
        ----------
        params : iterable
            Iterable of parameters to optimize or dicts defining parameter groups.
        lr : float, optional
            Learning rate. Default is ``1e-3``.
        betas : tuple of floats, optional
            Coefficients used for computing running averages of gradient and its square. Default is ``(0.9, 0.999)``.
        eps : float, optional
            Term added to the denominator to improve numerical stability. Default is ``1e-8``.
        weight_decay : float, optional
            Weight decay (L2 penalty). Default is ``0``.
        """
        defaults = {"lr": lr, "betas": betas, "eps": eps, "weight_decay": weight_decay}
        super().__init__(params, defaults)

    def step(self, closure=None):
        """Step through the optimizer"""
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError("RAdam does not support sparse gradients")

                p_data_fp32 = p.data
                if p.data.dtype in {torch.float16, torch.bfloat16}:
                    p_data_fp32 = p_data_fp32.float()

                state = self.state[p]

                # State Initialization
                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(grad)
                else:
                    state["exp_avg"] = state["exp_avg"].type_as(p_data_fp32)

                state["step"] += 1

                # Weight update
                exp_avg = state["exp_avg"]
                beta1, beta2 = group["betas"]

                step_size = exp_avg * beta1 + grad * (1 - beta1)

                # Decay the momentum running average coefficient
                exp_avg.mul_(beta2).add_(grad, alpha=1 - beta2)

                if group["weight_decay"] != 0:
                    p_data_fp32.add_(-group["weight_decay"] * group["lr"])

                # more conservative since it's an approximated value
                p_data_fp32.add_(torch.sign(step_size), alpha=-group["lr"])

                p.data.copy_(p_data_fp32)

            return loss



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NKI-AI/direct/blob/main/direct/nn/mwcnn/mwcnn.py

from collections import OrderedDict
from typing import Optional, Tuple

import torch
import torch.nn.functional as F
from torch import nn


class DWT(nn.Module):
    """2D Discrete Wavelet Transform as presented in [Liu2018]_.

    References
    ----------
    .. [Liu2018] Liu, Pengju, et al. “Multi-Level Wavelet-CNN for Image Restoration.” ArXiv:1805.07071 [Cs], May 2018.
        arXiv.org, http://arxiv.org/abs/1805.07071.
    """

    def __init__(self):
        """Inits :class:`DWT`."""
        super().__init__()
        self.requires_grad = False

    @staticmethod
    def forward(x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DWT`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            DWT of `x`.
        """
        x01 = x[:, :, 0::2, :] / 2
        x02 = x[:, :, 1::2, :] / 2
        x1 = x01[:, :, :, 0::2]
        x2 = x02[:, :, :, 0::2]
        x3 = x01[:, :, :, 1::2]
        x4 = x02[:, :, :, 1::2]
        x_LL = x1 + x2 + x3 + x4
        x_HL = -x1 - x2 + x3 + x4
        x_LH = -x1 + x2 - x3 + x4
        x_HH = x1 - x2 - x3 + x4

        return torch.cat((x_LL, x_HL, x_LH, x_HH), 1)


class IWT(nn.Module):
    """2D Inverse Wavelet Transform as presented in [Liu2018]_.

    References
    ----------
    .. [Liu2018] Liu, Pengju, et al. “Multi-Level Wavelet-CNN for Image Restoration.” ArXiv:1805.07071 [Cs], May 2018.
        arXiv.org, http://arxiv.org/abs/1805.07071.
    """

    def __init__(self):
        """Inits :class:`IWT`."""
        super().__init__()
        self.requires_grad = False
        self._r = 2

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`IWT`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            IWT of `x`.
        """
        batch, in_channel, in_height, in_width = x.size()
        out_channel, out_height, out_width = int(in_channel / (self._r**2)), self._r * in_height, self._r * in_width

        x1 = x[:, 0:out_channel, :, :] / 2
        x2 = x[:, out_channel : out_channel * 2, :, :] / 2
        x3 = x[:, out_channel * 2 : out_channel * 3, :, :] / 2
        x4 = x[:, out_channel * 3 : out_channel * 4, :, :] / 2

        h = torch.zeros([batch, out_channel, out_height, out_width], dtype=x.dtype).to(x.device)

        h[:, :, 0::2, 0::2] = x1 - x2 - x3 + x4
        h[:, :, 1::2, 0::2] = x1 - x2 + x3 - x4
        h[:, :, 0::2, 1::2] = x1 + x2 - x3 - x4
        h[:, :, 1::2, 1::2] = x1 + x2 + x3 + x4

        return h


class ConvBlock(nn.Module):
    """Convolution Block for MWCNN as presented in [Liu2018]_.

    References
    ----------
    .. [Liu2018] Liu, Pengju, et al. “Multi-Level Wavelet-CNN for Image Restoration.” ArXiv:1805.07071 [Cs], May 2018.
        arXiv.org, http://arxiv.org/abs/1805.07071.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        bias: bool = True,
        batchnorm: bool = False,
        activation: nn.Module = nn.ReLU(True),
        scale: Optional[float] = 1.0,
    ):
        """Inits :class:`ConvBlock`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        kernel_size : int
            Kernel size of convolution.
        bias : bool, optional
            Use convolution bias. Default is ``True``.
        batchnorm : bool, optional
            Use batch normalization. Default is ``False``.
        activation : torch.nn.Module, optional
            Activation function. Default is ``nn.ReLU(True)``.
        scale : float, optional
            Scale factor for convolution. Default is ``1.0``.
        """
        super().__init__()

        net = [
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                bias=bias,
                padding=kernel_size // 2,
            )
        ]

        if batchnorm:
            net.append(nn.BatchNorm2d(num_features=out_channels, eps=1e-4, momentum=0.95))
        net.append(activation)

        self.net = nn.Sequential(*net)
        self.scale = scale

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ConvBlock`.

        Parameters
        ----------
        x : torch.Tensor
            Input with shape (N, C, H, W).

        Returns
        -------
        torch.Tensor
            Output with shape (N, C', H', W').
        """
        return self.net(x) * self.scale


class DilatedConvBlock(nn.Module):
    """Double dilated Convolution Block fpr MWCNN as presented in [Liu2018]_.

    References
    ----------
    .. [Liu2018] Liu, Pengju, et al. “Multi-Level Wavelet-CNN for Image Restoration.” ArXiv:1805.07071 [Cs], May 2018.
        arXiv.org, http://arxiv.org/abs/1805.07071.
    """

    def __init__(
        self,
        in_channels: int,
        dilations: Tuple[int, int],
        kernel_size: int,
        out_channels: Optional[int] = None,
        bias: bool = True,
        batchnorm: bool = False,
        activation: nn.Module = nn.ReLU(True),
        scale: Optional[float] = 1.0,
    ):
        """Inits :class:`DilatedConvBlock`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        dilations : Tuple[int, int]
            Dilations for first and second convolution.
        kernel_size : int
            Kernel size of convolution.
        out_channels : int
            Number of output channels.
        bias : bool, optional
            Use convolution bias. Default is ``True``.
        batchnorm : bool, optional
            Use batch normalization. Default is ``False``.
        activation : torch.nn.Module, optional
            Activation function. Default is ``nn.ReLU(True)``.
        scale : float, optional
            Scale factor for convolution. Default is ``1.0``.
        """
        super().__init__()
        net = [
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=in_channels,
                kernel_size=kernel_size,
                bias=bias,
                dilation=dilations[0],
                padding=kernel_size // 2 + dilations[0] - 1,
            )
        ]

        if batchnorm:
            net.append(nn.BatchNorm2d(num_features=in_channels, eps=1e-4, momentum=0.95))
        net.append(activation)
        if out_channels is None:
            out_channels = in_channels
        net.append(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                bias=bias,
                dilation=dilations[1],
                padding=kernel_size // 2 + dilations[1] - 1,
            )
        )
        if batchnorm:
            net.append(nn.BatchNorm2d(num_features=in_channels, eps=1e-4, momentum=0.95))
        net.append(activation)

        self.net = nn.Sequential(*net)
        self.scale = scale

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DilatedConvBlock`.

        Parameters
        ----------
        x : torch.Tensor
            Input with shape (N, C, H, W).

        Returns
        -------
        torch.Tensor
            Output with shape (N, C', H', W').
        """
        return self.net(x) * self.scale


class MWCNN(nn.Module):
    r"""Multi-level Wavelet CNN (MWCNN) implementation as presented in [Liu2018]_.

    References
    ----------
    .. [Liu2018] Liu, Pengju, et al. “Multi-Level Wavelet-CNN for Image Restoration.” ArXiv:1805.07071 [Cs], May 2018.
        arXiv.org, http://arxiv.org/abs/1805.07071.

    """

    def __init__(
        self,
        input_channels: int,
        first_conv_hidden_channels: int,
        num_scales: int = 4,
        bias: bool = True,
        batchnorm: bool = False,
        activation: nn.Module = nn.ReLU(True),
    ):
        """Inits :class:`MWCNN`.

        Parameters
        ----------
        input_channels : int
            Number of input channels.
        first_conv_hidden_channels : int
            Number of hidden channels in first convolution.
        num_scales : int, optional
            Number of scales. Default is ``4``.
        bias : bool, optional
            Use convolution bias. Default is ``True``.
        batchnorm : bool, optional
            Use batch normalization. Default is ``False``.
        activation : torch.nn.Module, optional
            Activation function. Default is ``nn.ReLU(True)``.
        """
        super().__init__()
        self._kernel_size = 3
        self.DWT = DWT()
        self.IWT = IWT()

        self.down = nn.ModuleList()
        for idx in range(num_scales):
            in_channels = input_channels if idx == 0 else first_conv_hidden_channels * 2 ** (idx + 1)
            out_channels = first_conv_hidden_channels * 2**idx
            dilations = (2, 1) if idx != num_scales - 1 else (2, 3)
            self.down.append(
                nn.Sequential(
                    OrderedDict(
                        [
                            (
                                f"convblock{idx}",
                                ConvBlock(
                                    in_channels=in_channels,
                                    out_channels=out_channels,
                                    kernel_size=self._kernel_size,
                                    bias=bias,
                                    batchnorm=batchnorm,
                                    activation=activation,
                                ),
                            ),
                            (
                                f"dilconvblock{idx}",
                                DilatedConvBlock(
                                    in_channels=out_channels,
                                    dilations=dilations,
                                    kernel_size=self._kernel_size,
                                    bias=bias,
                                    batchnorm=batchnorm,
                                    activation=activation,
                                ),
                            ),
                        ]
                    )
                )
            )
        self.up = nn.ModuleList()
        for idx in range(num_scales)[::-1]:
            in_channels = first_conv_hidden_channels * 2**idx
            out_channels = input_channels if idx == 0 else first_conv_hidden_channels * 2 ** (idx + 1)
            dilations = (2, 1) if idx != num_scales - 1 else (3, 2)
            self.up.append(
                nn.Sequential(
                    OrderedDict(
                        [
                            (
                                f"invdilconvblock{num_scales - 2 - idx}",
                                DilatedConvBlock(
                                    in_channels=in_channels,
                                    dilations=dilations,
                                    kernel_size=self._kernel_size,
                                    bias=bias,
                                    batchnorm=batchnorm,
                                    activation=activation,
                                ),
                            ),
                            (
                                f"invconvblock{num_scales - 2 - idx}",
                                ConvBlock(
                                    in_channels=in_channels,
                                    out_channels=out_channels,
                                    kernel_size=self._kernel_size,
                                    bias=bias,
                                    batchnorm=batchnorm,
                                    activation=activation,
                                ),
                            ),
                        ]
                    )
                )
            )
        self.num_scales = num_scales

    @staticmethod
    def pad(x: torch.Tensor) -> torch.Tensor:
        """Pads input to height and width dimensions if odd.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Padded tensor.
        """
        padding = [0, 0, 0, 0]

        if x.shape[-2] % 2 != 0:
            padding[3] = 1  # Padding right - width
        if x.shape[-1] % 2 != 0:
            padding[1] = 1  # Padding bottom - height
        if sum(padding) != 0:
            x = F.pad(x, padding, "reflect")
        return x

    @staticmethod
    def crop_to_shape(x: torch.Tensor, shape: tuple) -> torch.Tensor:
        r"""Crops ``x`` to specified shape.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor with shape ('\'*, H, W).
        shape : tuple
            Crop shape corresponding to H, W.

        Returns
        -------
        torch.Tensor
            Cropped tensor.
        """
        h, w = x.shape[-2:]
        if h > shape[0]:
            x = x[:, :, : shape[0], :]
        if w > shape[1]:
            x = x[:, :, :, : shape[1]]
        return x

    def forward(self, input_tensor: torch.Tensor, res: bool = False) -> torch.Tensor:
        r"""Forward pass of :class:`MWCNN`.

        Parameters
        ----------
        input_tensor : torch.Tensor
            Input tensor with shape ('\'*, C, H, W).
        res : bool, optional
            If True, residual connection is applied to the output. Default is ``False``.

        Returns
        -------
        torch.Tensor
            Output tensor with shape ('\'*, C, H, W).
        """
        res_values = []
        x = self.pad(input_tensor.clone())
        for idx in range(self.num_scales):
            if idx == 0:
                x = self.pad(self.down[idx](x))
                res_values.append(x)
            elif idx == self.num_scales - 1:
                x = self.down[idx](self.DWT(x))
            else:
                x = self.pad(self.down[idx](self.DWT(x)))
                res_values.append(x)

        for idx in range(self.num_scales):
            if idx != self.num_scales - 1:
                x = (
                    self.crop_to_shape(self.IWT(self.up[idx](x)), res_values[self.num_scales - 2 - idx].shape[-2:])
                    + res_values[self.num_scales - 2 - idx]
                )
            else:
                x = self.crop_to_shape(self.up[idx](x), input_tensor.shape[-2:])
                if res:
                    x = x + input_tensor
        return x



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import List, Tuple, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts import coil_combination_method
from atommic.collections.multitask.rs.nn.base import BaseMRIReconstructionSegmentationModel
from atommic.collections.multitask.rs.nn.seranet_base.convlstm_unet import ConvLSTMNormUnet
from atommic.collections.multitask.rs.nn.seranet_base.seranet_block import (
    SERANetReconstructionBlock,
    SERANetRecurrentBlock,
)
from atommic.collections.reconstruction.nn.ccnn_base.ccnn_block import CascadeNetBlock, Conv2d
from atommic.collections.reconstruction.nn.unet_base.unet_block import Unet
from atommic.collections.segmentation.nn.attentionunet_base.attentionunet_block import AttentionGate
from atommic.core.classes.common import typecheck

__all__ = ["SERANet"]


class SERANet(BaseMRIReconstructionSegmentationModel):
    """Implementation of the End-to-End Recurrent Attention Network as presented in [Huang2019]_.

    References
    ----------
    .. [Huang2019] Huang, Q., Chen, X., Metaxas, D., Nadar, M.S. (2019). Brain Segmentation from k-Space with
       End-to-End Recurrent Attention Network. In: , et al. Medical Image Computing and Computer Assisted
       Intervention – MICCAI 2019. Lecture Notes in Computer Science(), vol 11766. Springer, Cham.
       https://doi.org/10.1007/978-3-030-32248-9_31
    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`SERANet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object.
        trainer : Trainer, optional
            PyTorch Lightning trainer object. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.input_channels = cfg_dict.get("input_channels", 2)
        if self.input_channels == 0:
            raise ValueError("Segmentation module input channels cannot be 0.")
        if self.input_channels > 2:
            raise ValueError(f"Segmentation module input channels must be either 1 or 2. Found: {self.input_channels}")
        self.consecutive_slices = cfg_dict.get("consecutive_slices", 1)

        reconstruction_module = cfg_dict.get("reconstruction_module", "unet")
        reconstruction_module_output_channels = cfg_dict.get("reconstruction_module_output_channels", 1)
        if reconstruction_module.lower() == "unet":
            regularizer = Unet(
                in_chans=self.input_channels,
                out_chans=reconstruction_module_output_channels,
                chans=cfg_dict.get("reconstruction_module_channels", 64),
                num_pool_layers=cfg_dict.get("reconstruction_module_pooling_layers", 2),
                drop_prob=cfg_dict.get("reconstruction_module_dropout", 0.0),
            )
        elif reconstruction_module.lower() == "cascadenet":
            regularizer = torch.nn.ModuleList(
                [
                    CascadeNetBlock(
                        Conv2d(
                            in_channels=self.input_channels,
                            out_channels=reconstruction_module_output_channels,
                            hidden_channels=cfg_dict.get("reconstruction_module_hidden_channels", 64),
                            n_convs=cfg_dict.get("reconstruction_module_n_convs", 2),
                            batchnorm=cfg_dict.get("reconstruction_module_batchnorm", True),
                        ),
                        fft_centered=self.fft_centered,
                        fft_normalization=self.fft_normalization,
                        spatial_dims=self.spatial_dims,
                        coil_dim=self.coil_dim if self.consecutive_slices == 1 else self.coil_dim - 1,
                        no_dc=True,
                    )
                    for _ in range(cfg_dict.get("reconstruction_module_num_cascades", 5))
                ]
            )
        else:
            raise ValueError(f"Unknown reconstruction module: {reconstruction_module} for SERANet")

        self.reconstruction_module = SERANetReconstructionBlock(
            num_reconstruction_blocks=cfg_dict.get("reconstruction_module_num_blocks", 3),
            reconstruction_model=regularizer,
            fft_centered=self.fft_centered,
            fft_normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
            coil_dim=self.coil_dim if self.consecutive_slices == 1 else self.coil_dim - 1,
            coil_combination_method=self.coil_combination_method,
        )
        self.segmentation_module_input_channels = cfg_dict.get("segmentation_module_input_channels", 2)
        segmentation_module_output_channels = cfg_dict.get("segmentation_module_output_channels", 1)
        self.segmentation_module = ConvLSTMNormUnet(
            in_chans=self.segmentation_module_input_channels,
            out_chans=segmentation_module_output_channels,
            chans=cfg_dict.get("segmentation_module_channels", 64),
            num_pools=cfg_dict.get("segmentation_module_pooling_layers", 2),
            drop_prob=cfg_dict.get("segmentation_module_dropout", 0.0),
        )
        self.recurrent_module = SERANetRecurrentBlock(
            num_iterations=cfg_dict.get("recurrent_module_iterations", 3),
            attention_model=AttentionGate(
                in_chans_x=self.segmentation_module_input_channels * 2,
                in_chans_g=segmentation_module_output_channels,
                out_chans=segmentation_module_output_channels,
            ),
            unet_model=ConvLSTMNormUnet(
                in_chans=self.segmentation_module_input_channels * 2,
                out_chans=segmentation_module_output_channels,
                chans=cfg_dict.get("recurrent_module_attention_channels", 64),
                num_pools=cfg_dict.get("recurrent_module_attention_pooling_layers", 2),
                drop_prob=cfg_dict.get("recurrent_module_attention_dropout", 0.0),
            ),
            fft_centered=self.fft_centered,
            fft_normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )

        self.magnitude_input = cfg_dict.get("magnitude_input", True)
        self.normalize_segmentation_output = cfg_dict.get("normalize_segmentation_output", True)

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        init_reconstruction_pred: torch.Tensor,
        target_reconstruction: torch.Tensor,  # pylint: disable=unused-argument
        hx: torch.Tensor = None,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> Tuple[Union[List, torch.Tensor], torch.Tensor]:
        """Forward pass of :class:`SERANet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        init_reconstruction_pred : torch.Tensor
            Initial reconstruction prediction. Shape [batch_size, n_x, n_y, 2]
        target_reconstruction : torch.Tensor
            Target reconstruction. Shape [batch_size, n_x, n_y, 2]
        hx : torch.Tensor, optional
            Initial hidden state for the RNN. Default is ``None``.
        sigma : float, optional
            Standard deviation of the noise. Default is ``1.0``.

        Returns
        -------
        Tuple[Union[List, torch.Tensor], torch.Tensor]
            Tuple containing the predicted reconstruction and segmentation.
        """
        if self.consecutive_slices > 1:
            batch, slices = init_reconstruction_pred.shape[:2]
            init_reconstruction_pred = init_reconstruction_pred.reshape(
                init_reconstruction_pred.shape[0] * init_reconstruction_pred.shape[1],
                *init_reconstruction_pred.shape[2:],
            )
            y = y.reshape(y.shape[0] * y.shape[1], *y.shape[2:])
            mask = mask.reshape(mask.shape[0] * mask.shape[1], *mask.shape[2:])
            sensitivity_maps = sensitivity_maps.reshape(
                sensitivity_maps.shape[0] * sensitivity_maps.shape[1],
                *sensitivity_maps.shape[2:],
            )

        if init_reconstruction_pred.shape[-1] == 2:
            if self.input_channels == 1:
                init_reconstruction_pred = torch.view_as_complex(init_reconstruction_pred).unsqueeze(1)
                if self.magnitude_input:
                    init_reconstruction_pred = torch.abs(init_reconstruction_pred)
            elif self.input_channels == 2:
                if self.magnitude_input:
                    raise ValueError("Magnitude input is not supported for 2-channel input.")
                init_reconstruction_pred = init_reconstruction_pred.permute(0, 3, 1, 2)
            else:
                raise ValueError(f"The input channels must be either 1 or 2. Found: {self.input_channels}")
        else:
            if init_reconstruction_pred.dim() == 3:
                init_reconstruction_pred = init_reconstruction_pred.unsqueeze(1)

        reconstruction = self.reconstruction_module(init_reconstruction_pred, y, sensitivity_maps, mask)

        if len(reconstruction) > 1:
            pred_reconstruction = reconstruction[-2]
        else:
            pred_reconstruction = reconstruction[-1]

        segmentation = reconstruction[-1]

        if segmentation.shape[-1] == 2:
            segmentation = torch.abs(torch.view_as_complex(segmentation))

        # In case of deviating number of coils, we need to pad up to maximum number of coils == number of input \
        # channels for the reconstruction module
        num_coils = segmentation.shape[1]
        if num_coils != self.segmentation_module_input_channels:
            num_coils_to_add = self.segmentation_module_input_channels - num_coils
            dummy_segmentation_coil_data = torch.zeros_like(
                torch.movedim(segmentation, self.coil_dim, 0)[0]
            ).unsqueeze(self.coil_dim)
            dummy_coil_data = torch.zeros_like(torch.movedim(pred_reconstruction, self.coil_dim, 0)[0]).unsqueeze(
                self.coil_dim
            )
            for _ in range(num_coils_to_add):
                segmentation = torch.cat([segmentation, dummy_segmentation_coil_data], dim=self.coil_dim)
                pred_reconstruction = torch.cat([pred_reconstruction, dummy_coil_data], dim=self.coil_dim)
                y = torch.cat([y, dummy_coil_data], dim=self.coil_dim)
                sensitivity_maps = torch.cat([sensitivity_maps, dummy_coil_data], dim=self.coil_dim)

        segmentation = self.segmentation_module(segmentation)

        pred_segmentation = self.recurrent_module(pred_reconstruction, segmentation, y, sensitivity_maps, mask)

        if self.normalize_segmentation_output:
            pred_segmentation = (pred_segmentation - pred_segmentation.min()) / (
                pred_segmentation.max() - pred_segmentation.min()
            )

        pred_segmentation = torch.abs(pred_segmentation)

        pred_reconstruction = coil_combination_method(
            pred_reconstruction,
            sensitivity_maps,
            method=self.coil_combination_method,
            dim=self.coil_dim if self.consecutive_slices == 1 else self.coil_dim - 1,
        )
        pred_reconstruction = torch.view_as_complex(pred_reconstruction)

        if self.consecutive_slices > 1:
            pred_reconstruction = pred_reconstruction.view([batch, slices, *pred_reconstruction.shape[1:]])
            pred_segmentation = pred_segmentation.view([batch, slices, *pred_segmentation.shape[1:]])

        return pred_reconstruction, pred_segmentation



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/callbacks/nemo_model_checkpoint.py

import os
import re
import shutil
from copy import deepcopy
from pathlib import Path
from typing import Iterable, Optional, Union

import pytorch_lightning
import torch
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.utilities import rank_zero_info

from atommic.collections.common.callbacks import EMA
from atommic.utils import logging, model_utils
from atommic.utils.app_state import AppState
from atommic.utils.get_rank import is_global_rank_zero


class ATOMMICModelCheckpoint(ModelCheckpoint):
    """Light wrapper around Lightning's ModelCheckpoint to force a saved checkpoint on train_end.
    Extends Lightning's on_save_checkpoint func to save the .atommic file. Saves the .atommic file based
    on the best checkpoint saved (according to the monitor value).
    Also contains func to save the EMA copy of the model.
    """

    def __init__(
        self,
        always_save_atommic: bool = False,
        save_atommic_on_train_end: bool = True,
        save_best_model: bool = False,
        postfix: str = ".atommic",
        n_resume: bool = False,
        model_parallel_size: int = None,
        **kwargs,
    ):
        """Inits :class:`ATOMMICModelCheckpoint`

        Parameters
        ----------
        always_save_atommic : bool, optional
            Whether to always save the .atommic file. Default is ``False``.
        save_atommic_on_train_end : bool, optional
            Whether to save the .atommic file on train end. Default is ``True``.
        save_best_model : bool, optional
            Whether to save the best model. Default is ``False``.
        postfix : str, optional
            The postfix of the .atommic file. Default is ``.atommic``.
        n_resume : bool, optional
            Whether to resume from a previous run. Default is ``False``.
        model_parallel_size : int, optional
            The model parallel size. Default is ``None``.
        """
        # Parse and store "extended" parameters: save_best model and postfix.
        self.always_save_atommic = always_save_atommic
        self.save_atommic_on_train_end = save_atommic_on_train_end
        self.save_best_model = save_best_model
        if self.save_best_model and not self.save_atommic_on_train_end:
            logging.warning(
                (
                    "Found save_best_model is True and save_atommic_on_train_end is False. "
                    "Set save_atommic_on_train_end to True to automatically save the best model."
                )
            )
        self.postfix = postfix
        self.previous_best_path = ""
        self.model_parallel_size = model_parallel_size

        # `prefix` is deprecated
        if "prefix" in kwargs:
            self.prefix = kwargs.pop("prefix")
        else:
            self.prefix = ""

        # Call the parent class constructor with the remaining kwargs.
        super().__init__(**kwargs)

        if self.save_top_k != -1 and n_resume:
            logging.debug("Checking previous runs")
            self.atommic_topk_check_previous_run()

    def atommic_topk_check_previous_run(self):
        """Checks if there are any previous runs and if so, loads the best model from the previous run."""
        try:
            self.best_k_models  # pylint: disable=pointless-statement
            self.kth_best_model_path  # pylint: disable=pointless-statement
            self.best_model_score  # pylint: disable=pointless-statement
            self.best_model_path  # pylint: disable=pointless-statement
        except AttributeError as e:
            raise AttributeError(
                "Lightning's ModelCheckpoint was updated. ATOMMICModelCheckpoint needs to update."
            ) from e
        self.best_k_models = {}
        self.kth_best_model_path = ""
        self.best_model_score = None
        self.best_model_path = ""

        checkpoints = list(path for path in self._saved_checkpoint_paths if not self._is_ema_filepath(path))
        for checkpoint in checkpoints:
            if "mp_rank" in str(checkpoint) or "tp_rank" in str(checkpoint):
                checkpoint = model_utils.uninject_model_parallel_rank(checkpoint)
            checkpoint = str(checkpoint)
            # second case is for distributed checkpoints, since they are a directory there's no extension
            if checkpoint[-10:] == '-last.ckpt' or checkpoint[-5:] == '-last':
                continue
            index = checkpoint.find(self.monitor) + len(self.monitor) + 1  # Find monitor in str + 1 for '='
            if index != len(self.monitor):
                match = re.search("[A-z]", checkpoint[index:])
                if match:
                    value = checkpoint[index : index + match.start() - 1]  # -1 due to separator hypen
                    self.best_k_models[checkpoint] = float(value)
        if len(self.best_k_models) < 1:
            return  # No saved checkpoints yet

        _reverse = bool(self.mode == "min")

        best_k_models = sorted(self.best_k_models, key=self.best_k_models.get, reverse=_reverse)

        # This section should be ok as rank zero will delete all excess checkpoints, since all other ranks are
        # instantiated after rank zero. models_to_delete should be 0 for all other ranks.
        if self.model_parallel_size is not None:
            # check for distributed checkpoint
            if checkpoints[0].is_dir():
                models_to_delete = len(best_k_models) - self.save_top_k
            else:
                models_to_delete = len(best_k_models) - self.model_parallel_size * self.save_top_k
        else:
            models_to_delete = len(best_k_models) - self.save_top_k

        models_to_delete = max(0, models_to_delete)
        logging.debug(f'Number of models to delete: {models_to_delete}')

        # If EMA enabled, delete the additional EMA weights
        ema_enabled = self._has_ema_ckpts(self._saved_checkpoint_paths)

        for _ in range(models_to_delete):
            model = best_k_models.pop(-1)
            self.best_k_models.pop(model)
            self._del_model_without_trainer(model)
            if ema_enabled and self._fs.exists(self._ema_format_filepath(model)):
                self._del_model_without_trainer(self._ema_format_filepath(model))
            logging.debug(f"Removed checkpoint: {model}")

        self.kth_best_model_path = best_k_models[-1]
        self.best_model_path = best_k_models[0]
        self.best_model_score = self.best_k_models[self.best_model_path]

    # pylint: disable=inconsistent-return-statements
    def on_save_checkpoint(self, trainer, pl_module, checkpoint):
        """Saves the .atommic file based on the best checkpoint saved (according to the monitor value)."""
        output = super().on_save_checkpoint(  # pylint: disable=assignment-from-no-return
            trainer, pl_module, checkpoint
        )
        if not self.always_save_atommic:
            return output
        # Load the best model and then re-save it
        app_state = AppState()
        if app_state.model_parallel_size is not None and app_state.model_parallel_size > 1:
            logging.warning("always_save_atommic will slow down training for model_parallel > 1.")
        # since we are creating tarfile artifacts we need to update .atommic path
        app_state.model_restore_path = os.path.abspath(
            os.path.expanduser(os.path.join(self.dirpath, self.prefix + self.postfix))
        )
        if app_state.model_parallel_size is not None and app_state.model_parallel_size > 1:
            maybe_injected_best_model_path = model_utils.inject_model_parallel_rank(self.best_model_path)
        else:
            maybe_injected_best_model_path = self.best_model_path

        if self.save_best_model:
            if not os.path.exists(maybe_injected_best_model_path):
                return

            if self.best_model_path == self.previous_best_path:
                return output

            old_state_dict = deepcopy(pl_module.state_dict())
            checkpoint = torch.load(maybe_injected_best_model_path, map_location="cpu")
            if "state_dict" in checkpoint:
                checkpoint = checkpoint["state_dict"]
            # get a new instanace of the model
            pl_module.load_state_dict(checkpoint, strict=True)
            if torch.distributed.is_initialized():
                torch.distributed.barrier()
            pl_module.save_to(save_path=app_state.model_restore_path)
            logging.info(f"New best .atommic model saved to: {app_state.model_restore_path}")
            pl_module.load_state_dict(old_state_dict, strict=True)
        else:
            if torch.distributed.is_initialized():
                torch.distributed.barrier()
            pl_module.save_to(save_path=app_state.model_restore_path)
            logging.info(f"New .atommic model saved to: {app_state.model_restore_path}")
        return output

    def on_train_end(self, trainer, pl_module):
        """Saves the .atommic file based on the best checkpoint saved (according to the monitor value)."""
        if trainer.fast_dev_run:
            return None

        # check if we need to save a last checkpoint manually as validation isn't always run based on the interval
        if self.save_last and trainer.val_check_interval != 0:
            should_save_last_checkpoint = False
            if isinstance(trainer.val_check_interval, float) and trainer.val_check_interval % trainer.global_step != 0:
                should_save_last_checkpoint = True
            if isinstance(trainer.val_check_interval, int) and trainer.global_step % trainer.val_check_interval != 0:
                should_save_last_checkpoint = True
            if should_save_last_checkpoint:
                monitor_candidates = self._monitor_candidates(trainer)
                super()._save_last_checkpoint(trainer, monitor_candidates)
        # Call parent on_train_end() to save the -last checkpoint
        super().on_train_end(trainer, pl_module)

        # Load the best model and then re-save it
        if self.save_best_model:
            # wait for all processes
            trainer.strategy.barrier("SaveBestCheckpointConnector.resume_end")
            if self.best_model_path == "":
                logging.warning(
                    f"{self} was told to save the best checkpoint at the end of training, but no saved checkpoints "
                    "were found. Saving latest model instead."
                )
            else:
                self.best_model_path = trainer.strategy.broadcast(self.best_model_path)
                trainer._checkpoint_connector.restore(self.best_model_path)  # pylint: disable=protected-access

        if self.save_atommic_on_train_end:
            pl_module.save_to(save_path=os.path.join(self.dirpath, self.prefix + self.postfix))

    def _del_model_without_trainer(self, filepath: str) -> None:
        """Deletes the checkpoint file without instantiating the model."""
        filepath = Path(filepath)  # type: ignore

        # check if filepath is a distributed a checkpoint
        if model_utils.ckpt_to_dir(filepath).is_dir():
            if is_global_rank_zero():
                try:
                    dist_ckpt = model_utils.ckpt_to_dir(filepath)
                    shutil.rmtree(dist_ckpt)
                    logging.info(f"Removed distributed checkpoint: {dist_ckpt}")
                except Exception:
                    logging.info(f"Tried to remove distributed checkpoint: {dist_ckpt} but failed.")

        else:
            app_state = AppState()

            # legacy model parallel checkpoint
            if app_state.model_parallel_size is not None and app_state.model_parallel_size > 1:
                # filepath needs to be updated to include mp_rank
                filepath = model_utils.inject_model_parallel_rank(filepath)

            # each model parallel rank needs to remove its model
            if is_global_rank_zero() or (
                app_state.model_parallel_size is not None and app_state.data_parallel_rank == 0
            ):
                try:
                    self._fs.rm(filepath)
                    logging.info(f"Removed checkpoint: {filepath}")
                except Exception:
                    logging.info(f"Tried to remove checkpoint: {filepath} but failed.")

    def _ema_callback(self, trainer: "pytorch_lightning.Trainer") -> Optional[EMA]:
        """Returns the EMA callback if it exists."""
        ema_callback = None
        for callback in trainer.callbacks:
            if isinstance(callback, EMA):
                ema_callback = callback
        return ema_callback

    def _save_checkpoint(self, trainer: "pytorch_lightning.Trainer", filepath: str) -> None:
        """Saves the checkpoint and the EMA copy of the model if EMA is enabled."""
        ema_callback = self._ema_callback(trainer)
        if ema_callback is not None:
            with ema_callback.save_original_optimizer_state(trainer):
                super()._save_checkpoint(trainer, filepath)

            # save EMA copy of the model as well.
            with ema_callback.save_ema_model(trainer):
                filepath = self._ema_format_filepath(filepath)
                if self.verbose:
                    rank_zero_info(f"Saving EMA weights to separate checkpoint {filepath}")
                super()._save_checkpoint(trainer, filepath)
        else:
            super()._save_checkpoint(trainer, filepath)

    def _remove_checkpoint(self, trainer: "pytorch_lightning.Trainer", filepath: str) -> None:
        """Removes the checkpoint and the EMA copy of the model if EMA is enabled."""
        super()._remove_checkpoint(trainer, filepath)
        ema_callback = self._ema_callback(trainer)
        if ema_callback is not None:
            # remove EMA copy of the state dict as well.
            filepath = self._ema_format_filepath(filepath)
            super()._remove_checkpoint(trainer, filepath)

    def _ema_format_filepath(self, filepath: str) -> str:
        """Returns the filepath of the EMA copy of the model."""
        return filepath.replace(self.FILE_EXTENSION, f"-EMA{self.FILE_EXTENSION}")

    def _has_ema_ckpts(self, checkpoints: Iterable[Path]) -> bool:
        """Returns True if any of the checkpoints are EMA checkpoints."""
        return any(self._is_ema_filepath(checkpoint_path) for checkpoint_path in checkpoints)

    def _is_ema_filepath(self, filepath: Union[Path, str]) -> bool:
        """Returns True if the filepath is an EMA checkpoint."""
        return str(filepath).endswith(f"-EMA{self.FILE_EXTENSION}")

    @property
    def _saved_checkpoint_paths(self) -> Iterable[Path]:
        """Returns the saved checkpoint paths."""
        # distributed checkpoints are directories so we check for them here
        dist_checkpoints = [d for d in list(Path(self.dirpath).glob("*")) if d.is_dir()]
        if dist_checkpoints:
            return dist_checkpoints
        return Path(self.dirpath).rglob("*.ckpt")



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/dynunet.py

from typing import List, Optional, Sequence, Tuple, Union

import torch
from torch import nn
from torch.nn.functional import interpolate

from atommic.collections.segmentation.nn.unetr_base.unetr_block import (
    UnetBasicBlock,
    UnetOutBlock,
    UnetResBlock,
    UnetUpBlock,
)

__all__ = ["DynUNet"]


class DynUNetSkipLayer(nn.Module):
    r"""Implementation of a Dynamic UNet (DynUNet) Skip Layer, based on [Isensee2018]_.

    References
    ----------
    .. [Isensee2018] Isensee F, Petersen J, Klein A, Zimmerer D, Jaeger PF, Kohl S, Wasserthal J, Koehler G,
        Norajitra T, Wirkert S, Maier-Hein KH. nnu-net: Self-adapting framework for u-net-based medical image
        segmentation. arXiv preprint arXiv:1809.10486. 2018 Sep 27.


    .. note::
        This class is a wrapper of the original DynUNetSkipLayer class from MONAI.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/dynunet.py


    .. note::
        Defines a layer in the UNet topology which combines the downsample and upsample pathways with the skip
        connection. The member `next_layer` may refer to instances of this class or the final bottleneck layer at the
        bottom the UNet structure. The purpose of using a recursive class like this is to get around the Torchscript
        restrictions on looping over lists of layers and accumulating lists of output tensors which must be indexed.
        The `heads` list is shared amongst all the instances of this class and is used to store the output from the
        supervision heads during forward passes of the network.

    """

    heads: Optional[List[torch.Tensor]]

    def __init__(
        self,
        index: int,
        downsample: nn.Module,
        upsample: nn.Module,
        next_layer: nn.Module,
        heads: List[torch.Tensor] = None,
        super_head: Optional[nn.Module] = None,
    ):
        """Inits :class:`DynUNetSkipLayer`.

         Parameters
         ----------
         index : int
             The index of the layer in the UNet structure.
         downsample : nn.Module
             The downsample layer of the skip connection.
         upsample : nn.Module
             The upsample layer of the skip connection.
         next_layer : nn.Module
             The next layer in the UNet structure.
         heads : List[torch.Tensor]
             The list of output tensors from the supervision heads. Default is ``None``.
        super_head : nn.Module
            The supervision head for this layer. Default is ``None``.
        """
        super().__init__()
        self.downsample = downsample
        self.next_layer = next_layer
        self.upsample = upsample
        self.super_head = super_head
        self.heads = heads
        self.index = index

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DynUNetSkipLayer`."""
        downout = self.downsample(x)
        nextout = self.next_layer(downout)
        upout = self.upsample(nextout, downout)
        if self.super_head is not None and self.heads is not None and self.index > 0:
            self.heads[self.index - 1] = self.super_head(upout)

        return upout


class DynUNet(nn.Module):
    """Implementation of a Dynamic UNet (DynUNet) Skip Layer, based on [Isensee2018]_.

    References
    ----------
    .. [Isensee2018] Isensee F, Petersen J, Klein A, Zimmerer D, Jaeger PF, Kohl S, Wasserthal J, Koehler G,
        Norajitra T, Wirkert S, Maier-Hein KH. nnu-net: Self-adapting framework for u-net-based medical image
        segmentation. arXiv preprint arXiv:1809.10486. 2018 Sep 27.

    .. note::
        This class is a wrapper of the original DynUNetSkipLayer class from MONAI.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/dynunet.py
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: Sequence[Union[Sequence[int], int]],
        strides: Sequence[Union[Sequence[int], int]],
        upsample_kernel_size: Sequence[Union[Sequence[int], int]],
        filters: Optional[Sequence[int]] = None,
        dropout: Optional[Union[Tuple, str, float]] = None,
        norm_name: Union[Tuple, str] = ("INSTANCE", {"affine": True}),
        act_name: Union[Tuple, str] = ("leakyrelu", {"inplace": True, "negative_slope": 0.01}),
        deep_supervision: bool = False,
        deep_supr_num: int = 1,
        res_block: bool = False,
        trans_bias: bool = False,
    ):
        """Inits :class:`DynUNet`.

        Parameters
        ----------
        spatial_dims : int
            The number of spatial dimensions of the input data.
        in_channels : int
            The number of input channels.
        out_channels : int
            The number of output channels.
        kernel_size : Union[int, Sequence[int]]
            The kernel size for the convolutional layers.
        strides : Union[int, Sequence[int]]
            The stride for the convolutional layers.
        upsample_kernel_size : Union[int, Sequence[int]]
            Convolution kernel size for transposed convolution layers. The values should equal to strides[1:].
        filters : Sequence[int]
            The number of output channels for each block. Different from nnU-Net, in this implementation we add this
            argument to make the network more flexible. One way to determine this parameter is like:
            ``[64, 96, 128, 192, 256, 384, 512, 768, 1024][: len(strides)]``. If not specified, the way which nnUNet
            used will be employed. Defaults to ``None``.
        dropout : float
            Dropout ratio. Defaults to no dropout.
        norm_name : str
            Feature normalization type and arguments. Defaults to ``INSTANCE``.
            `INSTANCE_NVFUSER` is a faster version of the instance norm layer, it can be used when:
            1) `spatial_dims=3`, 2) CUDA device is available, 3) `apex` is installed and 4) non-Windows OS is used.
        act_name : str
            Activation layer type and arguments. Defaults to ``leakyrelu``.
        deep_supervision : bool
            Whether to add deep supervision head before output. Defaults to ``False``. If ``True``, in training mode,
            the forward function will output not only the final feature map (from `output_block`), but also the feature
             maps that come from the intermediate up sample layers. In order to unify the return type (the restriction
             of TorchScript), all intermediate feature maps are interpolated into the same size as the final feature
             map and stacked together (with a new dimension in the first axis)into one single tensor. For instance, if
             there are two intermediate feature maps with shapes: (1, 2, 16, 12) and (1, 2, 8, 6), and the final
             feature map has the shape (1, 2, 32, 24), then all intermediate feature maps will be interpolated into
             (1, 2, 32, 24), and the stacked tensor will have the shape (1, 3, 2, 32, 24). When calculating the loss,
             you can use torch.unbind to get all feature maps can compute the loss one by one with the ground truth,
             then do a weighted average for all losses to achieve the final loss.
        deep_supr_num : int
            Number of feature maps that will output during deep supervision head. The value should be larger than 0 and
            less than the number of up sample layers. Defaults to ``1``.
        res_block : bool
            Whether to use residual connection based convolution blocks during the network. Defaults to ``False``.
        trans_bias : bool
            Whether to set the bias parameter in transposed convolution layers. Defaults to ``False``.
        """
        super().__init__()
        self.spatial_dims = spatial_dims
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.strides = strides
        self.upsample_kernel_size = upsample_kernel_size
        self.norm_name = norm_name
        self.act_name = act_name
        self.dropout = dropout
        self.conv_block = UnetResBlock if res_block else UnetBasicBlock
        self.trans_bias = trans_bias
        if filters is not None:
            self.filters = filters
            self.check_filters()
        else:
            self.filters = [min(2 ** (5 + i), 320 if spatial_dims == 3 else 512) for i in range(len(strides))]
        self.input_block = self.get_input_block()
        self.downsamples = self.get_downsamples()
        self.bottleneck = self.get_bottleneck()
        self.upsamples = self.get_upsamples()
        self.output_block = self.get_output_block(0)
        self.deep_supervision = deep_supervision
        self.deep_supr_num = deep_supr_num
        # initialize the typed list of supervision head outputs so that Torchscript can recognize what's going on
        self.heads: List[torch.Tensor] = [torch.rand(1)] * self.deep_supr_num
        if self.deep_supervision:
            self.deep_supervision_heads = self.get_deep_supervision_heads()
            self.check_deep_supr_num()

        self.apply(self.initialize_weights)
        self.check_kernel_stride()

        def create_skips(
            index: int,
            downsamples: List[nn.Module],
            upsamples: List[nn.Module],
            bottleneck: nn.Module,
            superheads: List[nn.Module] = None,
        ) -> nn.Module:
            """
            Construct the UNet topology as a sequence of skip layers terminating with the bottleneck layer. This is
            done recursively from the top down since a recursive nn.Module subclass is being used to be compatible
            with Torchscript. Initially the length of `downsamples` will be one more than that of `superheads`
            since the `input_block` is passed to this function as the first item in `downsamples`, however this
            shouldn't be associated with a supervision head.

            Parameters
            ----------
            index : int
                The index of the current skip layer.
            downsamples : List[nn.Module]
                The list of downsample layers.
            upsamples : List[nn.Module]
                The list of upsample layers.
            bottleneck : nn.Module
                The bottleneck layer.
            superheads : List[nn.Module]
                The list of supervision heads. Default is ``None``.
            """
            if len(downsamples) != len(upsamples):
                raise ValueError(f"{len(downsamples)} != {len(upsamples)}")

            if len(downsamples) == 0:  # bottom of the network, pass the bottleneck block
                return bottleneck

            if superheads is None:
                next_layer = create_skips(1 + index, downsamples[1:], upsamples[1:], bottleneck)
                return DynUNetSkipLayer(
                    index,
                    downsample=downsamples[0],
                    upsample=upsamples[0],
                    next_layer=next_layer,
                )

            super_head_flag = False
            if index == 0:  # don't associate a supervision head with self.input_block
                rest_heads = superheads
            else:
                if len(superheads) > 0:
                    super_head_flag = True
                    rest_heads = superheads[1:]
                else:
                    rest_heads = nn.ModuleList()

            # create the next layer down, this will stop at the bottleneck layer
            next_layer = create_skips(1 + index, downsamples[1:], upsamples[1:], bottleneck, superheads=rest_heads)
            if super_head_flag:
                return DynUNetSkipLayer(
                    index,
                    downsample=downsamples[0],
                    upsample=upsamples[0],
                    next_layer=next_layer,
                    heads=self.heads,
                    super_head=superheads[0],
                )

            return DynUNetSkipLayer(
                index,
                downsample=downsamples[0],
                upsample=upsamples[0],
                next_layer=next_layer,
            )

        if not self.deep_supervision:
            self.skip_layers = create_skips(
                0, [self.input_block] + list(self.downsamples), self.upsamples[::-1], self.bottleneck
            )
        else:
            self.skip_layers = create_skips(
                0,
                [self.input_block] + list(self.downsamples),
                self.upsamples[::-1],
                self.bottleneck,
                superheads=self.deep_supervision_heads,
            )

    def check_kernel_stride(self):
        """Check the length of kernel_size and strides."""
        kernels, strides = self.kernel_size, self.strides
        error_msg = "length of kernel_size and strides should be the same, and no less than 3."
        if len(kernels) != len(strides) or len(kernels) < 3:
            raise ValueError(error_msg)

        for idx, k_i in enumerate(kernels):
            kernel, stride = k_i, strides[idx]
            if not isinstance(kernel, int):
                error_msg = f"length of kernel_size in block {idx} should be the same as spatial_dims."
                if len(kernel) != self.spatial_dims:
                    raise ValueError(error_msg)
            if not isinstance(stride, int):
                error_msg = f"length of stride in block {idx} should be the same as spatial_dims."
                if len(stride) != self.spatial_dims:
                    raise ValueError(error_msg)

    def check_deep_supr_num(self):
        """Check the number of deep supervision heads."""
        deep_supr_num, strides = self.deep_supr_num, self.strides
        num_up_layers = len(strides) - 1
        if deep_supr_num >= num_up_layers:
            raise ValueError("deep_supr_num should be less than the number of up sample layers.")
        if deep_supr_num < 1:
            raise ValueError("deep_supr_num should be larger than 0.")

    def check_filters(self):
        """Check the length of filters."""
        filters = self.filters
        if len(filters) < len(self.strides):
            raise ValueError("Length of filters should be no less than the length of strides.")
        self.filters = filters[: len(self.strides)]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        out = self.skip_layers(x)
        out = self.output_block(out)
        if self.training and self.deep_supervision:
            out_all = [out]
            for feature_map in self.heads:
                out_all.append(interpolate(feature_map, out.shape[2:]))
            return torch.stack(out_all, dim=1)
        return out

    def get_input_block(self) -> nn.Module:
        """Get the input block."""
        return self.conv_block(
            self.spatial_dims,
            self.in_channels,
            self.filters[0],
            self.kernel_size[0],
            self.strides[0],
            self.norm_name,
            self.act_name,
            dropout=self.dropout,
        )

    def get_bottleneck(self) -> nn.Module:
        """Get the bottleneck block."""
        return self.conv_block(
            self.spatial_dims,
            self.filters[-2],
            self.filters[-1],
            self.kernel_size[-1],
            self.strides[-1],
            self.norm_name,
            self.act_name,
            dropout=self.dropout,
        )

    def get_output_block(self, idx: int) -> nn.Module:
        """Get the output block."""
        return UnetOutBlock(self.spatial_dims, self.filters[idx], self.out_channels, dropout=self.dropout)

    def get_downsamples(self) -> nn.ModuleList:
        """Get the downsampling blocks."""
        inp, out = self.filters[:-2], self.filters[1:-1]
        strides, kernel_size = self.strides[1:-1], self.kernel_size[1:-1]
        return self.get_module_list(inp, out, kernel_size, strides, self.conv_block)  # type: ignore

    def get_upsamples(self) -> nn.ModuleList:
        """Get the upsampling blocks."""
        inp, out = self.filters[1:][::-1], self.filters[:-1][::-1]
        strides, kernel_size = self.strides[1:][::-1], self.kernel_size[1:][::-1]
        upsample_kernel_size = self.upsample_kernel_size[::-1]
        return self.get_module_list(
            inp,  # type: ignore
            out,  # type: ignore
            kernel_size,
            strides,
            UnetUpBlock,
            upsample_kernel_size,
            trans_bias=self.trans_bias,
        )

    def get_module_list(
        self,
        in_channels: List[int],
        out_channels: List[int],
        kernel_size: Sequence[Union[Sequence[int], int]],
        strides: Sequence[Union[Sequence[int], int]],
        conv_block: nn.Module,
        upsample_kernel_size: Optional[Sequence[Union[Sequence[int], int]]] = None,
        trans_bias: bool = False,
    ) -> nn.ModuleList:
        """Get the module list of the network.

        Parameters
        ----------
        in_channels : List[int]
            The number of input channels.
        out_channels : List[int]
            The number of output channels.
        kernel_size : Sequence[Union[Sequence[int], int]]
            The kernel size.
        strides : Sequence[Union[Sequence[int], int]]
            The strides size.
        conv_block : nn.Module
            The convolutional block.
        upsample_kernel_size : Optional[Sequence[Union[Sequence[int], int]]]
            The upsample kernel size.
        trans_bias : bool
            Whether to use bias in the transpose convolutional layer.

        Returns
        -------
        nn.ModuleList
            The module list of the network.
        """
        layers = []
        if upsample_kernel_size is not None:
            for in_c, out_c, kernel, stride, up_kernel in zip(
                in_channels, out_channels, kernel_size, strides, upsample_kernel_size
            ):
                params = {
                    "spatial_dims": self.spatial_dims,
                    "in_channels": in_c,
                    "out_channels": out_c,
                    "kernel_size": kernel,
                    "stride": stride,
                    "norm_name": self.norm_name,
                    "act_name": self.act_name,
                    "dropout": self.dropout,
                    "upsample_kernel_size": up_kernel,
                    "trans_bias": trans_bias,
                }
                layer = conv_block(**params)
                layers.append(layer)
        else:
            for in_c, out_c, kernel, stride in zip(in_channels, out_channels, kernel_size, strides):
                params = {
                    "spatial_dims": self.spatial_dims,
                    "in_channels": in_c,
                    "out_channels": out_c,
                    "kernel_size": kernel,
                    "stride": stride,
                    "norm_name": self.norm_name,
                    "act_name": self.act_name,
                    "dropout": self.dropout,
                }
                layer = conv_block(**params)
                layers.append(layer)
        return nn.ModuleList(layers)

    def get_deep_supervision_heads(self) -> nn.ModuleList:
        """Get the deep supervision heads."""
        return nn.ModuleList([self.get_output_block(i + 1) for i in range(self.deep_supr_num)])

    @staticmethod
    def initialize_weights(module: nn.Module):
        """Initialize the weights of the network."""
        if isinstance(module, (nn.Conv3d, nn.Conv2d, nn.ConvTranspose3d, nn.ConvTranspose2d)):
            module.weight = nn.init.kaiming_normal_(module.weight, a=0.01)
            if module.bias is not None:
                module.bias = nn.init.constant_(module.bias, 0)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from torch import nn

from atommic.collections.reconstruction.nn.unet_base.unet_block import ConvBlock, TransposeConvBlock


class AttentionGate(nn.Module):
    """A Convolutional Block that consists of two convolution layers each followed by instance normalization,
    LeakyReLU, activation and dropout, as presented in [Oktay2018]_.

    References
    ----------
    .. [Oktay2018] O. Oktay, J. Schlemper, L.L. Folgoc, M. Lee, M. Heinrich, K. Misawa, K. Mori, S. McDonagh, N.Y.
        Hammerla, B. Kainz, B. Glocker, D. Rueckert. Attention U-Net: Learning Where to Look for the Pancreas. 2018.
        https://arxiv.org/abs/1804.03999
    """

    def __init__(self, in_chans_x: int, in_chans_g: int, out_chans: int):
        """Inits :class:`AttentionGate`.

        Parameters
        ----------
        in_chans_x : int
            Number of input channels of the input tensor `x`.
        in_chans_g : int
            Number of input channels of the input tensor `g`.
        out_chans : int
            Number of output channels.
        """
        super().__init__()
        self.in_chans_x = in_chans_x
        self.in_chans_g = in_chans_g
        self.out_chans = out_chans

        self.W_x = nn.Sequential(nn.Conv2d(self.in_chans_x, out_chans, kernel_size=2, padding=0, stride=2, bias=False))
        self.W_g = nn.Sequential(nn.Conv2d(self.in_chans_g, out_chans, kernel_size=1, padding=0, bias=True))
        self.psi = nn.Sequential(nn.Conv2d(self.out_chans, 1, kernel_size=1, padding=0, bias=True))

    def forward(self, x: torch.Tensor, g: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`AttentionGate`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor `x` with shape [batch_size, in_chans_x, n_x, n_y].
        g : torch.Tensor
            Input tensor `g` with shape [batch_size, in_chans_g, n_x, n_y].

        Returns
        -------
        torch.Tensor
            Output tensor with shape [batch_size, out_chans, n_x, n_y].
        """
        W_x = self.W_x(x)
        w_g = self.W_g(g)
        W_g = nn.functional.interpolate(w_g, size=(W_x.shape[-2], W_x.shape[-1]), mode="bilinear", align_corners=False)
        f = nn.functional.relu(W_x + W_g, inplace=True)
        a = torch.sigmoid(self.psi(f))
        a = nn.functional.interpolate(a, size=(x.shape[-2], x.shape[-1]), mode="bilinear", align_corners=False)
        return a * x


class AttentionUnet(nn.Module):
    """Implementation of the Attention UNet for MRI segmentation, as presented in [Oktay2018]_.

    References
    ----------
    .. [Oktay2018] O. Oktay, J. Schlemper, L.L. Folgoc, M. Lee, M. Heinrich, K. Misawa, K. Mori, S. McDonagh, N.Y.
        Hammerla, B. Kainz, B. Glocker, D. Rueckert. Attention U-Net: Learning Where to Look for the Pancreas. 2018.
        https://arxiv.org/abs/1804.03999
    """

    def __init__(
        self,
        in_chans: int,
        out_chans: int,
        chans: int = 32,
        num_pool_layers: int = 4,
        drop_prob: float = 0.0,
        block=ConvBlock,
        **kwargs,
    ):
        """Inits :class:`AttentionUnet`.

        Parameters
        ----------
        in_chans : int
            Number of input channels.
        out_chans : int
            Number of output channels.
        chans : int
            Number of channels in the convolutional layers.
        num_pool_layers : int
            Number of pooling layers.
        drop_prob : float
            Dropout probability.
        block : nn.Module
            Convolutional block to use.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.chans = chans
        self.num_pool_layers = num_pool_layers
        self.drop_prob = drop_prob

        self.down_sample_layers = nn.ModuleList([ConvBlock(in_chans, chans, drop_prob)])
        ch = chans
        for _ in range(num_pool_layers - 1):
            self.down_sample_layers.append(block(ch, ch * 2, drop_prob, **kwargs))
            ch = ch * 2
        self.conv = block(ch, ch * 2, drop_prob, **kwargs)

        self.up_conv = nn.ModuleList()
        self.up_transpose_conv = nn.ModuleList()
        self.up_attention_gates = nn.ModuleList()
        for _ in range(num_pool_layers - 1):
            self.up_transpose_conv.append(TransposeConvBlock(ch * 2, ch))
            self.up_conv.append(ConvBlock(ch * 2, ch, drop_prob))
            self.up_attention_gates.append(AttentionGate(ch, ch * 2, ch))
            ch = ch // 2

        self.up_transpose_conv.append(TransposeConvBlock(ch * 2, ch))
        self.up_conv.append(
            nn.Sequential(
                ConvBlock(ch * 2, ch, drop_prob, **kwargs),
                nn.Conv2d(ch, self.out_chans, kernel_size=1, stride=1),
            )
        )
        self.up_attention_gates.append(AttentionGate(ch, ch * 2, ch))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`AttentionUnet`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor with shape [batch_size, in_chans, n_x, n_y].

        Returns
        -------
        torch.Tensor
            Output tensor with shape [batch_size, out_chans, n_x, n_y].
        """
        stack = []
        output = x

        # apply down-sampling layers
        for layer in self.down_sample_layers:
            output = layer(output)
            stack.append(output)
            output = nn.functional.avg_pool2d(output, kernel_size=2, stride=2, padding=0)

        output = self.conv(output)

        # apply up-sampling layers
        for transpose_conv, conv, attention_gate in zip(self.up_transpose_conv, self.up_conv, self.up_attention_gates):
            downsample_layer = stack.pop()
            downsample_layer = attention_gate(downsample_layer, output)
            output = transpose_conv(output)

            # reflect pad on the right/bottom if needed to handle odd input dimensions
            padding = [0, 0, 0, 0]
            if output.shape[-1] != downsample_layer.shape[-1]:
                padding[1] = 1  # padding right
            if output.shape[-2] != downsample_layer.shape[-2]:
                padding[3] = 1  # padding bottom
            if torch.sum(torch.tensor(padding)) != 0:
                output = nn.functional.pad(output, padding, "reflect")

            output = torch.cat([output, downsample_layer], dim=1)
            output = conv(output)

        return output



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.common.callbacks.callbacks import LogEpochTimeCallback  # noqa: F401
from atommic.collections.common.callbacks.ema import EMA  # noqa: F401



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import math
from typing import List, Tuple

import torch

from atommic.collections.multitask.rs.nn.seranet_base.convlstm import ConvLSTM
from atommic.collections.reconstruction.nn.unet_base.unet_block import Unet


class ConvLSTMNormUnet(torch.nn.Module):
    """Normalized U-Net with additional Convolutional LSTM input layer model.

    This is the same as a regular U-Net, but with normalization applied to the input before the U-Net. This keeps the
    values more numerically stable during training.
    """

    def __init__(
        self,
        chans: int,
        num_pools: int,
        in_chans: int = 2,
        out_chans: int = 2,
        drop_prob: float = 0.0,
        padding_size: int = 15,
        normalize: bool = True,
        norm_groups: int = 2,
    ):
        """Inits :class:`ConvLSTMNormUnet`.

        Parameters
        ----------
        chans : int
            Number of output channels of the first convolution layer.
        num_pools : int
            Number of down-sampling and up-sampling layers.
        in_chans : int
            Number of channels in the input to the U-Net model. Default is ``2``.
        out_chans : int
            Number of channels in the output to the U-Net model. Default is ``2``.
        drop_prob : float
            Dropout probability. Default is ``0.0``.
        padding_size: int
            Size of the padding. Default is ``15``.
        normalize: bool
            Whether to normalize the input. Default is ``True``.
        norm_groups: int
            Number of groups to use for group normalization. Default is ``2``.
        """
        super().__init__()
        self.convlstm = ConvLSTM(in_chans, chans, kernel_size=3, num_layers=1)
        self.unet = Unet(
            in_chans=chans, out_chans=out_chans, chans=chans, num_pool_layers=num_pools, drop_prob=drop_prob
        )
        self.padding_size = padding_size
        self.normalize = normalize
        self.norm_groups = norm_groups

    @staticmethod
    def complex_to_chan_dim(x: torch.Tensor) -> torch.Tensor:
        """Convert the last dimension of the input to complex."""
        b, c, h, w, two = x.shape
        if two != 2:
            raise AssertionError
        return x.permute(0, 4, 1, 2, 3).reshape(b, 2 * c, h, w)

    @staticmethod
    def chan_complex_to_last_dim(x: torch.Tensor) -> torch.Tensor:
        """Convert the last dimension of the input to complex."""
        b, c2, h, w = x.shape
        if c2 % 2 != 0:
            raise AssertionError
        c = torch.div(c2, 2, rounding_mode="trunc")
        return x.view(b, 2, c, h, w).permute(0, 2, 3, 4, 1).contiguous()

    def norm(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Normalize the input."""
        # group norm
        b, c, h, w = x.shape

        x = x.reshape(b, self.norm_groups, -1)

        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)

        x = (x - mean) / std

        x = x.reshape(b, c, h, w)

        return x, mean, std

    def unnorm(self, x: torch.Tensor, mean: torch.Tensor, std: torch.Tensor) -> torch.Tensor:
        """Unnormalize the input."""
        b, c, h, w = x.shape
        input_data = x.reshape(b, self.norm_groups, -1)
        return (input_data * std + mean).reshape(b, c, h, w)

    def pad(self, x: torch.Tensor) -> Tuple[torch.Tensor, Tuple[List[int], List[int], int, int]]:
        """Pad the input with zeros to make it square."""
        _, _, h, w = x.shape
        w_mult = ((w - 1) | self.padding_size) + 1
        h_mult = ((h - 1) | self.padding_size) + 1
        w_pad = [math.floor((w_mult - w) / 2), math.ceil((w_mult - w) / 2)]
        h_pad = [math.floor((h_mult - h) / 2), math.ceil((h_mult - h) / 2)]
        x = torch.nn.functional.pad(x, w_pad + h_pad)

        return x, (h_pad, w_pad, h_mult, w_mult)

    @staticmethod
    def unpad(x: torch.Tensor, h_pad: List[int], w_pad: List[int], h_mult: int, w_mult: int) -> torch.Tensor:
        """Unpad the input."""
        return x[..., h_pad[0] : h_mult - h_pad[1], w_pad[0] : w_mult - w_pad[1]]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ConvLSTMNormUnet`."""
        iscomplex = False
        if x.shape[-1] == 2:
            x = self.complex_to_chan_dim(x)
            iscomplex = True

        mean = 1.0
        std = 1.0

        if self.normalize:
            x, mean, std = self.norm(x)

        x, pad_sizes = self.pad(x)

        x, _ = self.convlstm(x.unsqueeze(0))
        x = x[0]
        if x.shape[0] == 1:
            x = x.squeeze(0)
        elif x.shape[1] == 1:
            x = x.squeeze(1)
        else:
            raise AssertionError
        x = self.unet(x)
        x = self.unpad(x, *pad_sizes)

        if self.normalize:
            x = self.unnorm(x, mean, std)

        if iscomplex:
            x = self.chan_complex_to_last_dim(x)

        return x



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NKI-AI/direct/blob/main/direct/algorithms/mri_algorithms.py

from abc import ABC
from typing import Callable, Optional, Sequence

import numpy as np
import torch

from atommic.collections.common.parts.fft import ifft2
from atommic.collections.common.parts.utils import crop_to_acs


class MaximumEigenvaluePowerMethod(ABC):
    """A class for solving the maximum eigenvalue problem using the Power Method.

    The Power Method is an iterative algorithm that can be used to find the largest eigenvalue of a matrix. The
    algorithm is initialized with a random vector and iteratively updates the vector by multiplying it with the matrix
    and normalizing it. The algorithm converges to the eigenvector corresponding to the largest eigenvalue. The
    largest eigenvalue is then estimated by taking the dot product of the eigenvector with the matrix.
    """

    def __init__(
        self,
        forward_operator: Callable,
        norm_func: Optional[Callable] = None,
        max_iter: int = 30,
    ):
        """Inits :class:`MaximumEigenvaluePowerMethod`.

        Parameters
        ----------
        forward_operator : Callable
            The forward operator for the problem.
        norm_func : Callable, optional
            An optional function for normalizing the eigenvector. Default is ``None``.
        max_iter : int, optional
            Maximum number of iterations to run the algorithm. Default is ``30``.
        """
        super().__init__()
        self.forward_operator = forward_operator
        self.norm_func = norm_func
        self.max_iter = max_iter
        self.iter = 0

    def _update(self) -> None:
        """Perform a single update step of the maximum eigenvalue guess and corresponding eigenvector."""
        y = self.forward_operator(self.x)  # type: ignore
        if self.norm_func is None:
            self.max_eig = (y * self.x.conj()).sum() / (self.x * self.x.conj()).sum()  # type: ignore
        else:
            self.max_eig = self.norm_func(y)
        self.x = y / self.max_eig

    def _done(self) -> bool:
        """Check if the algorithm is done."""
        return self.iter >= self.max_iter

    def _fit(self, x: torch.Tensor) -> None:
        """Sets initial maximum eigenvector guess."""
        self.x = x

    def update(self) -> None:
        """Update the algorithm's parameters and increment the iteration count."""
        self._update()
        self.iter += 1

    def done(self) -> bool:
        """Check if the algorithm has converged."""
        return self._done()

    def fit(self, *args, **kwargs) -> None:
        """Fit the algorithm.

        Parameters
        ----------
        *args : tuple
            Tuple of arguments for `_fit` method.
        **kwargs : dict
            Keyword arguments for `_fit` method.
        """
        self._fit(*args, **kwargs)
        while not self.done():
            self.update()


class EspiritCalibration(ABC, torch.nn.Module):
    """Estimates sensitivity maps estimated with the ESPIRIT calibration method as described in [1]_.

    We adapted code for ESPIRIT method adapted from [2]_.

    References
    ----------

    .. [1] Uecker M, Lai P, Murphy MJ, Virtue P, Elad M, Pauly JM, Vasanawala SS, Lustig M. ESPIRiT--an eigenvalue
        approach to autocalibrating parallel MRI: where SENSE meets GRAPPA. Magn Reson Med. 2014 Mar;71(3):990-1001.
        doi: 10.1002/mrm.24751. PMID: 23649942; PMCID: PMC4142121.
    .. [2] https://github.com/mikgroup/sigpy/blob/1817ff849d34d7cbbbcb503a1b310e7d8f95c242/sigpy/mri/app.py#L388-L491
    """

    def __init__(
        self,
        threshold: float = 0.05,
        kernel_size: int = 6,
        crop: float = 0.95,
        max_iter: int = 100,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Sequence[int] = (-2, -1),
    ):
        """Inits :class:`EstimateSensitivityMap`.

        Parameters
        ----------
        threshold: float, optional
            Threshold for the calibration matrix. Default: 0.05.
        kernel_size: int, optional
            Kernel size for the calibration matrix. Default: 6.
        crop: float, optional
            Output eigenvalue cropping threshold. Default: 0.95.
        max_iter: int, optional
            Power method iterations. Default: 30.
        fft_centered: bool, optional
            Whether to center the FFT. Default is ``False``.
        fft_normalization: str, optional
            Normalization to apply to the FFT. Default is ``"backward"``.
        spatial_dims: Sequence[int], optional
            Spatial dimensions of the input. Default is ``(-2, -1)``.
        """
        super().__init__()

        self.threshold = threshold
        self.kernel_size = kernel_size
        self.crop = crop
        self.max_iter = max_iter
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims

    def calculate_sensitivity_map(self, acs_mask: torch.Tensor, kspace: torch.Tensor) -> torch.Tensor:
        """Calculates sensitivity map given as input the `acs_mask` and the `k-space`.

        Parameters
        ----------
        acs_mask : torch.Tensor
            Autocalibration mask.
        kspace : torch.Tensor
            K-space.

        Returns
        -------
        sensitivity_map : torch.Tensor
            Coil sensitivity maps.
        """
        ndim = kspace.ndim - 2
        spatial_size = kspace.shape[1:-1]

        # Used in case the k-space is padded (e.g. for batches)
        non_padded_dim = kspace.clone().sum(dim=tuple(range(1, kspace.ndim))).bool()

        num_coils = non_padded_dim.sum()
        acs_kspace_cropped = torch.view_as_complex(crop_to_acs(acs_mask.squeeze(), kspace[non_padded_dim]))

        # Get calibration matrix.
        calibration_matrix = (
            torch.nn.functional.unfold(acs_kspace_cropped.unsqueeze(0), kernel_size=self.kernel_size, stride=1)
            .transpose(1, 2)
            .to(acs_kspace_cropped.device)
            .reshape(
                num_coils,
                *(np.array(acs_kspace_cropped.shape[1:3]) - self.kernel_size + 1),
                *([self.kernel_size] * ndim),
            )
        )
        calibration_matrix = calibration_matrix.reshape(num_coils, -1, self.kernel_size**ndim)
        calibration_matrix = calibration_matrix.permute(1, 0, 2)
        calibration_matrix = calibration_matrix.reshape(-1, num_coils * self.kernel_size**ndim)

        _, s, vh = torch.linalg.svd(calibration_matrix, full_matrices=True)
        vh = torch.where(s > (self.threshold * s.max()), vh, torch.zeros_like(vh))

        # Get kernels
        num_kernels = vh.shape[0]
        kernels = vh.reshape([num_kernels, num_coils] + [self.kernel_size] * ndim)

        # Get covariance matrix in image domain
        covariance = torch.zeros(
            spatial_size[::-1] + (num_coils, num_coils),
            dtype=kernels.dtype,
            device=kernels.device,
        )
        for kernel in kernels:
            pad_h, pad_w = (
                spatial_size[0] - self.kernel_size,
                spatial_size[1] - self.kernel_size,
            )
            pad = (pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2)
            kernel_padded = torch.nn.functional.pad(kernel, pad)

            img_kernel = ifft2(kernel_padded, self.fft_centered, self.fft_normalization, self.spatial_dims)
            img_kernel = torch.view_as_complex(img_kernel)

            aH = img_kernel.permute(*torch.arange(img_kernel.ndim - 1, -1, -1)).unsqueeze(-1)
            a = aH.transpose(-1, -2).conj()
            covariance = covariance + aH @ a

        covariance = covariance * (np.prod(spatial_size) / self.kernel_size**ndim)
        sensitivity_map = torch.ones(
            (*spatial_size[::-1], num_coils, 1),
            dtype=kernels.dtype,
            device=kernels.device,
        )

        def forward(x):
            return covariance @ x

        def normalize(x):
            return (x.abs() ** 2).sum(dim=-2, keepdims=True) ** 0.5

        power_method = MaximumEigenvaluePowerMethod(forward, max_iter=self.max_iter, norm_func=normalize)
        power_method.fit(x=sensitivity_map)

        temp_sensitivity_map = power_method.x.squeeze(-1)
        temp_sensitivity_map = temp_sensitivity_map.permute(
            *torch.arange(temp_sensitivity_map.ndim - 1, -1, -1)
        ).squeeze(-1)
        temp_sensitivity_map = temp_sensitivity_map * temp_sensitivity_map.conj() / temp_sensitivity_map.abs()

        max_eig = power_method.max_eig.squeeze()
        max_eig = max_eig.permute(*torch.arange(max_eig.ndim - 1, -1, -1))
        temp_sensitivity_map = temp_sensitivity_map * (max_eig > self.crop)

        sensitivity_map = torch.zeros_like(kspace, device=kspace.device, dtype=kspace.dtype)
        sensitivity_map[non_padded_dim] = torch.view_as_real(temp_sensitivity_map)
        return sensitivity_map

    def forward(self, acs_mask: torch.Tensor, kspace: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`EspiritCalibration`.

        Parameters
        ----------
        acs_mask : torch.Tensor
            Autocalibration mask.
        kspace : torch.Tensor
            K-space.

        Returns
        -------
        sensitivity_map : torch.Tensor
            Coil sensitivity maps.
        """
        return self.calculate_sensitivity_map(acs_mask, kspace)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

ATOMMIC_ENV_VARNAME_ENABLE_COLORING = "ATOMMIC_ENABLE_COLORING"
ATOMMIC_ENV_VARNAME_REDIRECT_LOGS_TO_STDERR = "ATOMMIC_REDIRECT_LOGS_TO_STDERR"
# Set to True to enable atommic.util.logging's debug mode
ATOMMIC_ENV_VARNAME_TESTING = "ATOMMIC_TESTING"
# Used for atommic.utils.exp_manager versioning
ATOMMIC_ENV_VARNAME_VERSION = "ATOMMIC_EXPM_VERSION"
# Used to change default atommic cache directory
ATOMMIC_ENV_CACHE_DIR = "ATOMMIC_CACHE_DIR"
# Used to change default atommic data store cache directory
ATOMMIC_ENV_DATA_STORE_CACHE_DIR = "ATOMMIC_DATA_STORE_CACHE_DIR"
# Shared among nodes (1) or not shared (0)
ATOMMIC_ENV_DATA_STORE_CACHE_SHARED = "ATOMMIC_DATA_STORE_CACHE_SHARED"



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Parts of the code have been taken from https://github.com/facebookresearch/fastMRI

import torch
import torch.nn.functional as F

from atommic.core.classes.loss import Loss


class SSIMLoss(Loss):
    """Computes the (1-) SSIM loss between two tensors.

    Examples
    --------
    >>> from atommic.collections.reconstruction.losses.ssim import SSIMLoss
    >>> import torch
    >>> loss = SSIMLoss(win_size=7, k1=0.01, k2=0.03)
    >>> loss(X=torch.rand(1, 1, 256, 256), Y=torch.rand(1, 1, 256, 256), data_range=torch.tensor([1.]))
    tensor(0.9872)
    """

    def __init__(self, win_size: int = 7, k1: float = 0.01, k2: float = 0.03):
        """Inits :class:`SSIMLoss`.

        Parameters
        ----------
        win_size : int, optional
            Window size for SSIM calculation.
        k1 : float, optional
            k1 parameter for SSIM calculation.
        k2 : float, optional
            k2 parameter for SSIM calculation.
        """
        super().__init__()
        self.win_size = win_size
        self.k1, self.k2 = k1, k2
        self.register_buffer("w", torch.ones(1, 1, win_size, win_size) / win_size**2)
        NP = win_size**2
        self.cov_norm = NP / (NP - 1)

    def forward(self, X: torch.Tensor, Y: torch.Tensor, data_range: torch.Tensor = None):
        """Forward pass of :class:`SSIMLoss`.

        Parameters
        ----------
        X : torch.Tensor
            First input tensor.
        Y : torch.Tensor
            Second input tensor.
        data_range : torch.Tensor
            Data range of the input tensors. If ``None``, it is computed as the maximum range of the input tensors.
            Default is ``None``.
        """
        if not isinstance(self.w, torch.Tensor):  # type: ignore  # pylint: disable=access-member-before-definition
            raise AssertionError

        # This is necessary to first assign self.w to CUDA and then in case of fp32 to avoid RuntimeError: Inference
        # tensors cannot be saved for backward.
        self.w = self.w.to(Y).clone()  # type: ignore

        if data_range is None:
            data_range = torch.tensor([max(X.max() - X.min(), Y.max() - Y.min())]).to(Y)
        if isinstance(data_range, int):
            data_range = torch.tensor([data_range]).to(Y)

        data_range = data_range[:, None, None, None]
        C1 = (self.k1 * data_range) ** 2
        C2 = (self.k2 * data_range) ** 2
        ux = F.conv2d(X, self.w)
        uy = F.conv2d(Y, self.w)
        uxx = F.conv2d(X * X, self.w)
        uyy = F.conv2d(Y * Y, self.w)
        uxy = F.conv2d(X * Y, self.w)
        vx = self.cov_norm * (uxx - ux * ux)
        vy = self.cov_norm * (uyy - uy * uy)
        vxy = self.cov_norm * (uxy - ux * uy)
        A1, A2, B1, B2 = (2 * ux * uy + C1, 2 * vxy + C2, ux**2 + uy**2 + C1, vx + vy + C2)
        D = B1 * B2
        S = (A1 * A2) / D

        return 1 - S.mean()



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.package_info import (  # noqa: F401
    __contact_emails__,
    __contact_names__,
    __description__,
    __download_url__,
    __homepage__,
    __keywords__,
    __license__,
    __package_name__,
    __repository_url__,
    __shortversion__,
    __version__,
)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/Project-MONAI/MONAI/blob/dev/monai/metrics/utils.py

from typing import Any, Tuple

import torch
from torch import Tensor


def do_metric_reduction(f: torch.Tensor, reduction: str = "mean") -> Tuple[Tensor, Any]:
    """Utility function to perform metric reduction.

    Parameters
    ----------
    f : torch.Tensor
        the metric to reduce.
    reduction : str
        the reduction method, default is ``mean``.

    Returns
    -------
    torch.Tensor or Any
        the reduced metric.
    Any
        NaNs if there are any NaNs in the input, otherwise 0.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.segmentation.losses.utils import do_metric_reduction
    >>> f = torch.tensor([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]])
    >>> do_metric_reduction(f, "mean")
    (tensor(6.5000), 0)
    >>> do_metric_reduction(f, "sum")
    (tensor(78), 0)
    """
    # some elements might be Nan (if ground truth y was missing (zeros)), we need to account for it
    nans = torch.isnan(f)
    not_nans = (~nans).float()
    t_zero = torch.zeros(1, device=f.device, dtype=f.dtype)
    if reduction is None:
        return f, not_nans
    f[nans] = 0
    if reduction == "mean":
        # 2 steps, first, mean by channel (accounting for nans), then by batch
        not_nans = not_nans.sum(dim=1)
        f = torch.where(not_nans > 0, f.sum(dim=1) / not_nans, t_zero)  # channel average
        not_nans = (not_nans > 0).float().sum(dim=0)
        f = torch.where(not_nans > 0, f.sum(dim=0) / not_nans, t_zero)  # batch average
    elif reduction == "sum":
        not_nans = not_nans.sum(dim=[0, 1])
        f = torch.sum(f, dim=[0, 1])  # sum over the batch and channel dims
    elif reduction == "mean_batch":
        not_nans = not_nans.sum(dim=0)
        f = torch.where(not_nans > 0, f.sum(dim=0) / not_nans, t_zero)  # batch average
    elif reduction == "sum_batch":
        not_nans = not_nans.sum(dim=0)
        f = f.sum(dim=0)  # the batch sum
    elif reduction == "mean_channel":
        not_nans = not_nans.sum(dim=1)
        f = torch.where(not_nans > 0, f.sum(dim=1) / not_nans, t_zero)  # channel average
    elif reduction == "sum_channel":
        not_nans = not_nans.sum(dim=1)
        f = f.sum(dim=1)  # the channel sum
    elif reduction == "none":
        pass
    else:
        raise ValueError(
            f"Unsupported reduction: {reduction}, available options are "
            '["mean", "sum", "mean_batch", "sum_batch", "mean_channel", "sum_channel" "none"].'
        )
    return f, not_nans



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import ifft2
from atommic.collections.common.parts.utils import check_stacked_complex, coil_combination_method
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.multidomainnet_base.multidomainnet_block import (
    MultiDomainUnet2d,
    StandardizationLayer,
)
from atommic.core.classes.common import typecheck

__all__ = ["MultiDomainNet"]


class MultiDomainNet(BaseMRIReconstructionModel):
    """Feature-level multi-domain module. Inspired by AIRS Medical submission to the FastMRI 2020 challenge."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`MultiDomainNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.num_cascades = cfg_dict.get("num_cascades")

        standardization = cfg_dict["standardization"]
        if standardization:
            self.standardization = StandardizationLayer(self.coil_dim, -1)

        self.reconstruction_module = MultiDomainUnet2d(
            # if standardization, in_channels is 4 due to standardized input
            in_channels=4 if standardization else 2,
            out_channels=2,
            num_filters=cfg_dict["num_filters"],
            num_pool_layers=cfg_dict["num_pool_layers"],
            dropout_probability=cfg_dict["dropout_probability"],
            fft_centered=self.fft_centered,
            fft_normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
            coil_dim=self.coil_dim,
        )

    def _compute_model_per_coil(self, model: torch.nn.Module, data: torch.Tensor) -> torch.Tensor:
        """Computes the model per coil.

        Parameters
        ----------
        model : torch.nn.Module
            The model to be computed.
        data : torch.Tensor
            The data to be computed. Shape [batch_size, n_coils, n_x, n_y, 2].

        Returns
        -------
        torch.Tensor
            The computed output. Shape [batch_size, n_coils, n_x, n_y, 2].
        """
        output = []
        for idx in range(data.size(self.coil_dim)):
            subselected_data = data.select(self.coil_dim, idx)
            output.append(model(subselected_data))
        output = torch.stack(output, dim=self.coil_dim)
        return output

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,  # pylint: disable=unused-argument
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`MultiDomainNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        image = ifft2(y, self.fft_centered, self.fft_normalization, self.spatial_dims)
        if hasattr(self, "standardization"):
            image = self.standardization(image, sensitivity_maps)
        prediction = self._compute_model_per_coil(self.reconstruction_module, image.permute(0, 1, 4, 2, 3)).permute(
            0, 1, 3, 4, 2
        )
        return check_stacked_complex(
            coil_combination_method(prediction, sensitivity_maps, self.coil_combination_method, self.coil_dim)
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NKI-AI/direct/blob/main/direct/nn/recurrentvarnet/recurrentvarnet.py

from typing import List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import complex_conj, complex_mul


class Conv2dGRU(nn.Module):
    """2D Convolutional Gated Recurrent Unit."""

    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        out_channels: Optional[int] = None,
        num_layers: int = 2,
        gru_kernel_size=1,
        orthogonal_initialization: bool = True,
        instance_norm: bool = False,
        dense_connect: int = 0,
        replication_padding: bool = True,
    ):
        """Inits :class:`Conv2dGRU`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        hidden_channels : int
            Number of hidden channels.
        out_channels : int, optional
            Number of output channels. If None, same as in_channels.
        num_layers : int, optional
            Number of layers. Default is ``2``.
        gru_kernel_size : int, optional
            Size of the GRU kernel. Default is ``1``.
        orthogonal_initialization : bool, optional
            Orthogonal initialization is used if set to True. Default is ``True``.
        instance_norm : bool, optional
            Instance norm is used if set to True. Default is ``False``.
        dense_connect : int, optional
            Number of dense connections. Default is ``0``.
        replication_padding : bool, optional
            If set to true replication padding is applied. Default is ``True``.
        """
        super().__init__()

        if out_channels is None:
            out_channels = in_channels

        self.num_layers = num_layers
        self.hidden_channels = hidden_channels
        self.dense_connect = dense_connect

        self.reset_gates = nn.ModuleList([])
        self.update_gates = nn.ModuleList([])
        self.out_gates = nn.ModuleList([])
        self.conv_blocks = nn.ModuleList([])

        # Create convolutional blocks
        for idx in range(num_layers + 1):
            in_ch = in_channels if idx == 0 else (1 + min(idx, dense_connect)) * hidden_channels
            out_ch = hidden_channels if idx < num_layers else out_channels
            padding = 0 if replication_padding else (2 if idx == 0 else 1)
            block = []
            if replication_padding:
                if idx == 1:
                    block.append(nn.ReplicationPad2d(2))
                else:
                    block.append(nn.ReplicationPad2d(2 if idx == 0 else 1))
            block.append(
                nn.Conv2d(
                    in_channels=in_ch,
                    out_channels=out_ch,
                    kernel_size=5 if idx == 0 else 3,
                    dilation=(2 if idx == 1 else 1),
                    padding=padding,
                )
            )
            self.conv_blocks.append(nn.Sequential(*block))

        # Create GRU blocks
        for _ in range(num_layers):
            for gru_part in [self.reset_gates, self.update_gates, self.out_gates]:
                block = []
                if instance_norm:
                    block.append(nn.InstanceNorm2d(2 * hidden_channels))
                block.append(
                    nn.Conv2d(
                        in_channels=2 * hidden_channels,
                        out_channels=hidden_channels,
                        kernel_size=gru_kernel_size,
                        padding=gru_kernel_size // 2,
                    )
                )
                gru_part.append(nn.Sequential(*block))

        if orthogonal_initialization:
            for reset_gate, update_gate, out_gate in zip(self.reset_gates, self.update_gates, self.out_gates):
                nn.init.orthogonal_(reset_gate[-1].weight)
                nn.init.orthogonal_(update_gate[-1].weight)
                nn.init.orthogonal_(out_gate[-1].weight)
                nn.init.constant_(reset_gate[-1].bias, -1.0)
                nn.init.constant_(update_gate[-1].bias, 0.0)
                nn.init.constant_(out_gate[-1].bias, 0.0)

    def forward(
        self,
        cell_input: torch.Tensor,
        previous_state: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass of :class:`Conv2dGRU`.

        Parameters
        ----------
        cell_input : torch.Tensor
            Input tensor.
        previous_state : torch.Tensor
            Previous hidden state.

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor]
            Tuple of output tensor and new hidden state.
        """
        new_states: List[torch.Tensor] = []
        conv_skip: List[torch.Tensor] = []

        if previous_state is None:
            batch_size, spatial_size = cell_input.size(0), (cell_input.size(2), cell_input.size(3))
            state_size = [batch_size, self.hidden_channels] + list(spatial_size) + [self.num_layers]
            previous_state = torch.zeros(*state_size, dtype=cell_input.dtype).to(cell_input.device)

        for idx in range(self.num_layers):
            if conv_skip:
                cell_input = F.relu(
                    self.conv_blocks[idx](torch.cat([*conv_skip[-self.dense_connect :], cell_input], dim=1)),
                    inplace=True,
                )
            else:
                cell_input = F.relu(self.conv_blocks[idx](cell_input), inplace=True)
            if self.dense_connect > 0:
                conv_skip.append(cell_input)

            stacked_inputs = torch.cat([cell_input, previous_state[:, :, :, :, idx]], dim=1)

            update = torch.sigmoid(self.update_gates[idx](stacked_inputs))
            reset = torch.sigmoid(self.reset_gates[idx](stacked_inputs))
            delta = torch.tanh(
                self.out_gates[idx](torch.cat([cell_input, previous_state[:, :, :, :, idx] * reset], dim=1))
            )
            cell_input = previous_state[:, :, :, :, idx] * (1 - update) + delta * update
            new_states.append(cell_input)
            cell_input = F.relu(cell_input, inplace=False)
        if conv_skip:
            out = self.conv_blocks[self.num_layers](torch.cat([*conv_skip[-self.dense_connect :], cell_input], dim=1))
        else:
            out = self.conv_blocks[self.num_layers](cell_input)

        return out, torch.stack(new_states, dim=-1)


class RecurrentInit(nn.Module):
    """Recurrent State Initializer (RSI) module of Recurrent Variational Network as presented in [Yiasemis2021]_.

    References
    ----------
    .. [Yiasemis2021] Yiasemis, George, et al. “Recurrent Variational Network: A Deep Learning Inverse Problem Solver
        Applied to the Task of Accelerated MRI Reconstruction.” ArXiv:2111.09639 [Physics], Nov. 2021. arXiv.org,
        http://arxiv.org/abs/2111.09639.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        channels: Tuple[int, ...],
        dilations: Tuple[int, ...],
        depth: int = 2,
        multiscale_depth: int = 1,
    ):
        """Inits :class:`RecurrentInit`.

        Parameters
        ----------
        in_channels : int
            Input channels.
        out_channels : int
            Number of hidden channels of the recurrent unit of RecurrentVarNet Block.
        channels : Tuple[int, ...]
            Channels :math:`n_d` in the convolutional layers of initializer.
        dilations : Tuple[int, ...]
            Dilations :math:`p` of the convolutional layers of the initializer.
        depth : int, optional
            RecurrentVarNet Block number of layers :math:`n_l`. Default is ``2``.
        multiscale_depth : int, optional
            Number of feature layers to aggregate for the output, if 1, multi-scale context aggregation is disabled.
            Default is ``1``.
        """
        super().__init__()

        self.conv_blocks = nn.ModuleList()
        self.out_blocks = nn.ModuleList()
        self.depth = depth
        self.multiscale_depth = multiscale_depth
        tch = in_channels
        for curr_channels, curr_dilations in zip(channels, dilations):
            block = [
                nn.ReplicationPad2d(curr_dilations),
                nn.Conv2d(tch, curr_channels, 3, padding=0, dilation=curr_dilations),
            ]
            tch = curr_channels
            self.conv_blocks.append(nn.Sequential(*block))
        tch = np.sum(channels[-multiscale_depth:])
        for _ in range(depth):
            block = [nn.Conv2d(tch, out_channels, 1, padding=0)]
            self.out_blocks.append(nn.Sequential(*block))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`RecurrentInit`.

        Parameters
        ----------
        x : torch.Tensor
            Initialization for RecurrentInit.

        Returns
        -------
        torch.Tensor
            Initial recurrent hidden state from input `x`.
        """
        features = []
        for block in self.conv_blocks:
            x = F.relu(block(x), inplace=True)
            if self.multiscale_depth > 1:
                features.append(x)
        if self.multiscale_depth > 1:
            x = torch.cat(features[-self.multiscale_depth :], dim=1)
        output_list = []
        for block in self.out_blocks:
            y = F.relu(block(x), inplace=True)
            output_list.append(y)
        return torch.stack(output_list, dim=-1)


class RecurrentVarNetBlock(nn.Module):
    r"""Recurrent Variational Network Block :math:`'\'mathcal{H}_{'\'theta_{t}}` as presented in [Yiasemis2021]_.

    References
    ----------
    .. [Yiasemis2021] Yiasemis, George, et al. “Recurrent Variational Network: A Deep Learning Inverse Problem Solver
        Applied to the Task of Accelerated MRI Reconstruction.” ArXiv:2111.09639 [Physics], Nov. 2021. arXiv.org,
        http://arxiv.org/abs/2111.09639.
    """

    def __init__(
        self,
        in_channels: int = 2,
        hidden_channels: int = 64,
        num_layers: int = 4,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
    ):
        """Inits :class:`RecurrentVarNetBlock`.

        Parameters
        ----------
        in_channels : int
            Input channels. Default is ``2`` for complex data.
        hidden_channels : int
            Number of hidden channels of the recurrent unit of RecurrentVarNet Block. Default is ``64``.
        num_layers : int
            Number of layers of :math:`n_l` recurrent unit. Default is ``4``.
        fft_centered : bool
            Whether to center the FFT. Default is ``False``.
        fft_normalization : str
            Whether to normalize the FFT. Default is ``"backward"``.
        spatial_dims : Tuple[int, int], optional
            Spatial dimensions of the input. Default is ``None``.
        coil_dim : int
            Coil dimension of the input. Default is ``1``.
        """
        super().__init__()
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim

        self.learning_rate = nn.Parameter(torch.tensor([1.0]))  # :math:`\alpha_t`
        self.regularizer = Conv2dGRU(
            in_channels=in_channels,
            hidden_channels=hidden_channels,
            num_layers=num_layers,
            replication_padding=True,
        )  # Recurrent Unit of RecurrentVarNet Block :math:`\mathcal{H}_{\theta_t}`

    def forward(
        self,
        current_kspace: torch.Tensor,
        masked_kspace: torch.Tensor,
        sampling_mask: torch.Tensor,
        sensitivity_map: torch.Tensor,
        hidden_state: Union[None, torch.Tensor],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass of :class:`RecurrentVarNetBlock`.

        Parameters
        ----------
        current_kspace : torch.Tensor
            Current k-space prediction. Shape [batch_size, n_coil, height, width, 2].
        masked_kspace : torch.Tensor
            Subsampled k-space. Shape [batch_size, n_coil, height, width, 2].
        sampling_mask : torch.Tensor
            Sampling mask. Shape [batch_size, 1, height, width, 1].
        sensitivity_map : torch.Tensor
            Coil sensitivities. Shape [batch_size, n_coil, height, width, 2].
        hidden_state : Union[None, torch.Tensor]
            ConvGRU hidden state. Shape [batch_size, n_l, height, width, hidden_channels].

        Returns
        -------
        new_kspace : torch.Tensor
            New k-space prediction. Shape [batch_size, n_coil, height, width, 2].
        new_hidden_state : list of torch.Tensor
            New ConvGRU hidden state. Shape [batch_size, n_l, height, width, hidden_channels].
        """
        kspace_error = torch.where(
            sampling_mask == 0,
            torch.tensor([0.0], dtype=masked_kspace.dtype).to(masked_kspace.device),
            current_kspace - masked_kspace,
        )

        recurrent_term = torch.cat(
            [
                complex_mul(
                    ifft2(
                        kspace,
                        centered=self.fft_centered,
                        normalization=self.fft_normalization,
                        spatial_dims=self.spatial_dims,
                    ),
                    complex_conj(sensitivity_map),
                ).sum(self.coil_dim)
                for kspace in torch.split(current_kspace, 2, -1)
            ],
            dim=-1,
        ).permute(0, 3, 1, 2)

        recurrent_term, hidden_state = self.regularizer(recurrent_term, hidden_state)  # :math:`w_t`, :math:`h_{t+1}`
        recurrent_term = recurrent_term.permute(0, 2, 3, 1)

        recurrent_term = torch.cat(
            [
                fft2(
                    complex_mul(image.unsqueeze(self.coil_dim), sensitivity_map),
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                )
                for image in torch.split(recurrent_term, 2, -1)
            ],
            dim=-1,
        )

        new_kspace = current_kspace - self.learning_rate * kspace_error + recurrent_term

        return new_kspace, hidden_state
